import io
from unittest.mock import Mock, mock_open, patch
import json
import pytest

from boxsdk.config import API
from boxsdk.object.email_alias import EmailAlias
from boxsdk.object.folder import Folder
from boxsdk.object.storage_policy_assignment import StoragePolicyAssignment
from boxsdk.network.default_network import DefaultNetworkResponse


def test_user_url(mock_user):
    # pylint:disable=redefined-outer-name, protected-access
    assert mock_user.get_url() == f'{API.BASE_API_URL}/users/{mock_user.object_id}'


def test_get_storage_policy_assignments(test_storage_policy_assignment, mock_user, mock_box_session):
    expected_url = mock_box_session.get_url('storage_policy_assignments')
    mock_assignment = {
        'type': test_storage_policy_assignment.object_type,
        'id': test_storage_policy_assignment.object_id,
        'assigned_to': {
            'type': mock_user.object_type,
            'id': mock_user.object_id,
        },
    }
    mock_box_session.get.return_value.json.return_value = {
        'next_marker': None,
        'limit': 1,
        'entries': [mock_assignment],
    }
    expected_params = {
        'resolved_for_type': mock_user.object_type,
        'resolved_for_id': mock_user.object_id,
    }
    assignment = mock_user.get_storage_policy_assignment()
    mock_box_session.get.assert_called_once_with(expected_url, params=expected_params)
    assert isinstance(assignment, StoragePolicyAssignment)
    assert assignment.id == mock_assignment['id']
    assert assignment.type == mock_assignment['type']


@pytest.fixture(scope='module')
def memberships_response():
    # pylint disable=redefined-outer-name
    mock_network_response = Mock(DefaultNetworkResponse)
    mock_network_response.json.return_value = {
        'entries': [
            {'type': 'group_membership', 'id': 101, 'user': {'type': 'user', 'id': 100},
             'group': {'type': 'group', 'id': 300}},
            {'type': 'group_membership', 'id': 202, 'user': {'type': 'user', 'id': 200},
             'group': {'type': 'group', 'id': 400}}
        ],
        'limit': 2,
        'total_count': 2,
        'offset': 0,
    }
    return mock_network_response


@pytest.fixture()
def test_email_alias(mock_box_session):
    return EmailAlias(
        session=mock_box_session,
        object_id='test_alias_id',
    )


def test_update(mock_user, mock_box_session):
    # pylint:disable=redefined-outer-name, protected-access
    user_id = mock_user.object_id
    expected_url = mock_box_session.get_url('users', user_id)
    user = {
        'type': 'user',
        'name': 'Test User',
        'id': 1234,
    }
    data = {
        'name': 'New User',
    }
    mock_box_session.put.return_value.json.return_value = user
    new_user = mock_user.update_info(data=data)
    mock_box_session.put.assert_called_once_with(expected_url, data=json.dumps(data), headers=None, params=None)
    assert new_user.id == user['id']
    assert new_user.type == user['type']
    assert new_user.name == user['name']


def test_delete(mock_user, mock_box_session):
    user_id = mock_user.object_id
    expected_url = mock_box_session.get_url('users', user_id)
    expected_params = {
        'notify': True,
        'force': False,
    }
    mock_user.delete()
    mock_box_session.delete.assert_called_once_with(
        expected_url,
        expect_json_response=False,
        params=expected_params,
        headers=None,
    )


def test_get_email_aliases(mock_user, mock_box_session):
    # pylint:disable=redefined-outer-name
    alias1_json = {
        'type': 'email_alias',
        'id': '12345',
        'email': 'foo@example.com',
    }
    alias2_json = {
        'type': 'email_alias',
        'id': '67890',
        'email': 'bar@example.com',
    }
    mock_box_session.get.return_value.json.return_value = {
        'total_count': 2,
        'entries': [alias1_json, alias2_json],
    }
    aliases = mock_user.get_email_aliases()
    for alias, alias_json in zip(aliases, [alias1_json, alias2_json]):
        assert alias.object_id == alias_json['id']
        # pylint:disable=protected-access
        assert alias._session == mock_box_session
        assert alias.email == alias_json['email']


def test_add_email_alias_returns_the_correct_email_alias_object(mock_user, mock_box_session):
    # pylint:disable=redefined-outer-name
    test_email_alias = 'test@example.com'
    expected_url = f'{API.BASE_API_URL}/users/{mock_user.object_id}/email_aliases'
    expected_body = json.dumps({
        'email': test_email_alias,
    })
    mock_box_session.post.return_value.json.return_value = {
        'type': 'email_alias',
        'id': '1234',
        'email': test_email_alias,
    }
    new_email_alias = mock_user.add_email_alias(test_email_alias)
    mock_box_session.post.assert_called_once_with(expected_url, data=expected_body)
    assert isinstance(new_email_alias, EmailAlias)
    assert new_email_alias._session == mock_box_session  # pylint: disable=protected-access
    assert new_email_alias.object_id == '1234'


def test_remove_email_alias(mock_user, mock_box_session, test_email_alias):
    expected_url = f'{API.BASE_API_URL}/users/{mock_user.object_id}/email_aliases/{test_email_alias.object_id}'
    mock_box_session.delete.return_value.ok = True

    result = mock_user.remove_email_alias(test_email_alias)

    mock_box_session.delete.assert_called_once_with(expected_url, expect_json_response=False)
    assert result is True


@pytest.mark.parametrize('notify,fields,expected_params', [
    (None, None, {}),
    (True, None, {'notify': True}),
    (False, None, {'notify': False}),
    (None, ['type', 'id', 'name'], {'fields': 'type,id,name'}),
    (False, ['type', 'id'], {'notify': False, 'fields': 'type,id'}),
])
def test_transfer_content(mock_user, mock_box_session, notify, fields, expected_params):
    # pylint:disable=redefined-outer-name
    expected_url = f"{API.BASE_API_URL}/users/{mock_user.object_id}/folders/0"
    expected_body = json.dumps({
        'owned_by': {
            'id': mock_user.object_id
        },
    })
    move_items_response = {
        'type': 'folder',
        'id': '12345',
        'name': 'That Other User\'s Content',
    }
    mock_box_session.put.return_value.json.return_value = move_items_response
    moved_item = mock_user.transfer_content(mock_user, notify=notify, fields=fields)
    mock_box_session.put.assert_called_once_with(expected_url, data=expected_body, params=expected_params)
    assert isinstance(moved_item, Folder)
    assert moved_item.id == move_items_response['id']
    assert moved_item.name == move_items_response['name']
    assert moved_item._session == mock_box_session  # pylint:disable=protected-access


def test_get_group_memberships(
        mock_user,
        mock_box_session,
        memberships_response,
):
    # pylint:disable=redefined-outer-name
    expected_url = f'{API.BASE_API_URL}/users/{mock_user.object_id}/memberships'
    mock_box_session.get.return_value = memberships_response
    memberships = mock_user.get_group_memberships()
    for membership, expected_id in zip(memberships, [101, 202]):
        assert membership.object_id == expected_id
        # pylint:disable=protected-access
        assert membership._session == mock_box_session
    mock_box_session.get.assert_called_once_with(expected_url, params={'offset': None})


def test_get_user_avatar(mock_user, mock_box_session, mock_content_response):
    expected_url = mock_user.get_url('avatar')
    mock_box_session.get.return_value = mock_content_response
    avatar_content = mock_user.get_avatar()
    assert avatar_content == mock_content_response.content
    mock_box_session.get.assert_called_once_with(
        expected_url,
        expect_json_response=False,
    )


@pytest.mark.parametrize('image_extension', ['jpg', 'jpeg', 'png'])
def test_upload_avatar(mock_user, mock_box_session, mock_image_path, image_extension):
    expected_url = mock_user.get_url('avatar')
    upload_avatar_response = {'pic_urls': {'large': 'url1', 'preview': 'url2', 'small': 'url3'}}
    mock_box_session.post.return_value.json.return_value = upload_avatar_response

    with patch('boxsdk.object.user.open', mock_open()) as mock_image:
        avatar_urls = mock_user.upload_avatar(mock_image_path)

        assert avatar_urls == upload_avatar_response['pic_urls']
        mock_box_session.post.assert_called_once_with(
            expected_url,
            files={'pic': (f'avatar.{image_extension}', mock_image.return_value, f'image/{image_extension}')}
        )


@pytest.mark.parametrize('image_extension', ['jpg', 'jpeg', 'png'])
def test_upload_avatar_stream(mock_user, mock_box_session, image_extension):
    expected_url = mock_user.get_url('avatar')
    upload_avatar_response = {'pic_urls': {'large': 'url1', 'preview': 'url2', 'small': 'url3'}}
    mock_box_session.post.return_value.json.return_value = upload_avatar_response
    image_stream = io.BytesIO(b"some image stream")

    avatar_urls = mock_user.upload_avatar_stream(image_stream=image_stream, image_extension=image_extension)

    assert avatar_urls == upload_avatar_response['pic_urls']
    mock_box_session.post.assert_called_once_with(
        expected_url,
        files={'pic': (f'avatar.{image_extension}', image_stream, f'image/{image_extension}')}
    )


def test_delete_avatar_stream(mock_user, mock_box_session):
    expected_url = mock_user.get_url('avatar')
    mock_box_session.delete.return_value.ok = True

    was_deleted_successfully = mock_user.delete_avatar()

    assert was_deleted_successfully
    mock_box_session.delete.assert_called_once_with(expected_url)



from datetime import datetime
from typing import Optional, Any, Union

from ..util.api_call_decorator import api_call
from ..util.default_arg_value import SDK_VALUE_NOT_SET
from .base_item import BaseItem


class WebLink(BaseItem):
    """Box API endpoint for interacting with WebLinks."""

    _item_type = 'web_link'

    @api_call
    def create_shared_link(
            self,
            *,
            access: Optional[str] = None,
            unshared_at: Union[datetime, str, None] = SDK_VALUE_NOT_SET,
            password: Optional[str] = None,
            vanity_name: Optional[str] = None,
            **kwargs: Any
    ) -> 'WebLink':
        """
        Baseclass override.

        :param access:
            Determines who can access the shared link. May be open, company, or collaborators. If no access is
            specified, the default access will be used.
        :param unshared_at:
            The date on which this link should be disabled. May only be set if the current user is not a free user
            and has permission to set expiration dates. Takes a datetime string supported by the dateutil library
            or a datetime.datetime object. If no timezone info provided, local timezone will be applied.
            The time portion can be omitted,
            which defaults to midnight (00:00:00) on that date.
        :param password:
            The password required to view this link. If no password is specified then no password will be set.
            Please notice that this is a premium feature, which might not be available to your app.
        :param vanity_name:
            Defines a custom vanity name to use in the shared link URL, eg. https://app.box.com/v/my-custom-vanity-name.
            If this parameter is None, the standard shared link URL will be used.
        :param kwargs:
            Used to fulfill the contract of overriden method
        :return:
            The updated object with s shared link.
            Returns a new object of the same type, without modifying the original object passed as self.
        """
        # pylint:disable=arguments-differ
        return super().create_shared_link(
            access=access,
            unshared_at=unshared_at,
            password=password,
            vanity_name=vanity_name
        )

    @api_call
    def get_shared_link(
            self,
            *,
            access: Optional[str] = None,
            unshared_at: Union[datetime, str, None] = SDK_VALUE_NOT_SET,
            password: Optional[str] = None,
            vanity_name: Optional[str] = None,
            **kwargs: Any
    ) -> str:
        """
        Baseclass override.

        :param access:
            Determines who can access the shared link. May be open, company, or collaborators. If no access is
            specified, the default access will be used.
        :param unshared_at:
            The date on which this link should be disabled. May only be set if the current user is not a free user
            and has permission to set expiration dates. Takes a datetime string supported by the dateutil library
            or a datetime.datetime object. If no timezone info provided, local timezone will be applied.
            The time portion can be omitted, which defaults to midnight (00:00:00) on that date.
        :param password:
            The password required to view this link. If no password is specified then no password will be set.
            Please notice that this is a premium feature, which might not be available to your app.
        :param vanity_name:
            Defines a custom vanity name to use in the shared link URL, eg. https://app.box.com/v/my-custom-vanity-name.
            If this parameter is None, the standard shared link URL will be used.
        :param kwargs:
            Used to fulfill the contract of overriden method
        :returns:
            The URL of the shared link.
        """
        # pylint:disable=arguments-differ
        return super().get_shared_link(
            access=access,
            unshared_at=unshared_at,
            password=password,
            vanity_name=vanity_name
        )

    @api_call
    def remove_shared_link(self, **kwargs: Any) -> bool:
        """
        Baseclass override.

        :param kwargs:
            Used to fulfill the contract of overriden method
        :returns:
            Whether or not the update was successful.
        """
        return super().remove_shared_link()



import json
import os
from datetime import datetime
from typing import TYPE_CHECKING, Optional, Tuple, Union, IO, Iterable, List, Any

from boxsdk.util.datetime_formatter import normalize_date_to_rfc3339_format
from .item import Item
from ..util.api_call_decorator import api_call
from ..util.default_arg_value import SDK_VALUE_NOT_SET
from ..util.deprecation_decorator import deprecated
from ..pagination.marker_based_object_collection import MarkerBasedObjectCollection
from ..pagination.limit_offset_based_object_collection import LimitOffsetBasedObjectCollection

if TYPE_CHECKING:
    from boxsdk.object.upload_session import UploadSession
    from boxsdk.util.chunked_uploader import ChunkedUploader
    from boxsdk.object.file_version import FileVersion
    from boxsdk.pagination.box_object_collection import BoxObjectCollection
    from boxsdk.object.comment import Comment
    from boxsdk.object.task import Task
    from boxsdk.object.folder import Folder


class File(Item):
    """Box API endpoint for interacting with files."""

    _item_type = 'file'

    @api_call
    def preflight_check(self, size: int, name: Optional[str] = None) -> Optional[str]:
        """
        Make an API call to check if the file can be updated with the new name and size of the file.
        Returns an accelerator URL if one is available.

        :param size:
            The size of the file in bytes. Specify 0 for unknown file-sizes.
        :param name:
            The name of the file to be updated. It's optional, if the name is not being changed.
        :return:
            The Accelerator upload url or None if cannot get the Accelerator upload url.
        :raises:
            :class:`BoxAPIException` when preflight check fails.
        """
        return self._preflight_check(
            size=size,
            name=name,
            file_id=self._object_id,
        )

    @api_call
    def create_upload_session(self, file_size: int, file_name: Optional[str] = None) -> 'UploadSession':
        """
        Create a new chunked upload session for uploading a new version of the file.

        :param file_size:
            The size of the file in bytes that will be uploaded.
        :param file_name:
            The new name of the file version that will be uploaded.
        :returns:
            A :class:`UploadSession` object.
        """
        body_params = {
            'file_id': self.object_id,
            'file_size': file_size,
        }
        if file_name is not None:
            body_params['file_name'] = file_name
        url = self.get_url('upload_sessions').replace(self.session.api_config.BASE_API_URL, self.session.api_config.UPLOAD_URL)
        response = self._session.post(url, data=json.dumps(body_params)).json()
        return self.translator.translate(
            session=self._session,
            response_object=response,
        )

    @api_call
    def get_chunked_uploader(self, file_path: str, rename_file: bool = False) -> 'ChunkedUploader':
        # pylint: disable=consider-using-with
        """
        Instantiate the chunked upload instance and create upload session with path to file.

        :param file_path:
            The local path to the file you wish to upload.
        :param rename_file:
            Indicates whether the file should be renamed or not.
        :returns:
            A :class:`ChunkedUploader` object.
        """
        total_size = os.stat(file_path).st_size
        content_stream = open(file_path, 'rb')
        file_name = os.path.basename(file_path) if rename_file else None
        upload_session = self.create_upload_session(total_size, file_name)
        return upload_session.get_chunked_uploader_for_stream(content_stream, total_size)

    def _get_accelerator_upload_url_for_update(self) -> Optional[str]:
        """
        Get Accelerator upload url for updating the file.

        :return:
            The Accelerator upload url for updating the file or None if cannot get one
        """
        return self._get_accelerator_upload_url(file_id=self._object_id)

    @staticmethod
    def _construct_range_header(boundaries: Union[Tuple[int], Tuple[int, int]]) -> str:
        """
        Construct the correct value for the Range header, given a closed or open-ended range.

        :param boundaries:
            The range of bytes (inclusive)
        :returns:
            The value for the Range header
        :raises ValueError:
        """
        if len(boundaries) == 1:
            return f'bytes={boundaries[0]}-'
        if len(boundaries) == 2:
            return f'bytes={boundaries[0]}-{boundaries[1]}'

        raise ValueError('Expected a 1-tuple or 2-tuple for byte range')

    @api_call
    def content(self, file_version: Optional['FileVersion'] = None, byte_range: Tuple[int, int] = None) -> bytes:
        """
        Get the content of a file on Box.

        :param file_version:
            The specific version of the file to retrieve the contents of.
        :param byte_range:
            A tuple of inclusive byte offsets to download, e.g. (100, 199) to download the second 100 bytes of a file
        :returns:
            File content as bytes.
        """
        url = self.get_url('content')
        params = {'version': file_version.object_id} if file_version is not None else None
        headers = {'Range': self._construct_range_header(byte_range)} if byte_range is not None else None
        box_response = self._session.get(url, expect_json_response=False, params=params, headers=headers)
        return box_response.content

    @api_call
    def download_to(
            self,
            writeable_stream: IO[bytes],
            file_version: Optional['FileVersion'] = None,
            byte_range: Tuple[int, int] = None
    ) -> None:
        """
        Download the file; write it to the given stream.

        :param writeable_stream:
            A file-like object where bytes can be written into.
        :param file_version:
            The specific version of the file to retrieve the contents of.
        :param byte_range:
            A tuple of inclusive byte offsets to download, e.g. (100, 199) to download the second 100 bytes of a file
        """
        url = self.get_url('content')
        params = {'version': file_version.object_id} if file_version is not None else None
        headers = {'Range': self._construct_range_header(byte_range)} if byte_range is not None else None
        box_response = self._session.get(url, expect_json_response=False, stream=True, params=params, headers=headers)
        for chunk in box_response.network_response.response_as_stream.stream(decode_content=True):
            writeable_stream.write(chunk)

    @api_call
    def get_download_url(self, file_version: Optional['FileVersion'] = None) -> str:
        """
        Get the url to download the file.

        :param file_version:
            The specific version of the file to retrieve the contents of.
        :return: Url to download the file
        """
        url = self.get_url('content')
        params = {'version': file_version.object_id} if file_version is not None else None
        box_response = self._session.get(
            url,
            params=params,
            expect_json_response=False,
            allow_redirects=False,
        )
        return box_response.headers['location']

    @api_call
    def update_contents_with_stream(
            self,
            file_stream: IO[bytes],
            etag: Optional[str] = None,
            preflight_check: bool = False,
            preflight_expected_size: int = 0,
            upload_using_accelerator: bool = False,
            file_name: Optional[str] = None,
            content_modified_at: Union[datetime, str] = None,
            additional_attributes: Optional[dict] = None,
            sha1: Optional[str] = None,
    ) -> 'File':
        """
        Upload a new version of a file, taking the contents from the given file stream.

        :param file_stream:
            The file-like object containing the bytes
        :param etag:
            If specified, instruct the Box API to update the item only if the current version's etag matches.
        :param preflight_check:
            If specified, preflight check will be performed before actually uploading the file.
        :param preflight_expected_size:
            The size of the file to be uploaded in bytes, which is used for preflight check. The default value is '0',
            which means the file size is unknown.
        :param upload_using_accelerator:
            If specified, the upload will try to use Box Accelerator to speed up the uploads for big files.
            It will make an extra API call before the actual upload to get the Accelerator upload url, and then make
            a POST request to that url instead of the default Box upload url. It falls back to normal upload endpoint,
            if cannot get the Accelerator upload url.

            Please notice that this is a premium feature, which might not be available to your app.
        :param file_name:
            The new name to give the file on Box.
        :param content_modified_at:
            The A datetime string in a format supported by the dateutil library or datetime object,
            which specifies when the file content was last modified.
            If no timezone info provided, local timezone will be applied.
        :param additional_attributes:
            A dictionary containing attributes to add to the file that are not covered by other parameters.
        :param sha1:
            A sha1 checksum for the new content.
        :returns:
            A new file object
        :raises:
            :class:`BoxAPIException` if the specified etag doesn't match the latest version of the file or preflight
            check fails.
        """
        accelerator_upload_url = None
        if preflight_check:
            # Preflight check does double duty, returning the accelerator URL if one is available in the response.
            accelerator_upload_url = self.preflight_check(size=preflight_expected_size)
        elif upload_using_accelerator:
            accelerator_upload_url = self._get_accelerator_upload_url_for_update()

        url = self.get_url('content').replace(
            self._session.api_config.BASE_API_URL,
            self._session.api_config.UPLOAD_URL,
        )
        if upload_using_accelerator and accelerator_upload_url:
            url = accelerator_upload_url

        attributes = {
            'name': file_name,
            'content_modified_at': normalize_date_to_rfc3339_format(content_modified_at),
        }
        if additional_attributes:
            attributes.update(additional_attributes)

        data = {'attributes': json.dumps(attributes)}
        files = {'file': ('unused', file_stream)}
        headers = {}
        if etag is not None:
            headers['If-Match'] = etag
        if sha1 is not None:
            # The Content-MD5 field accepts sha1
            headers['Content-MD5'] = sha1
        if not headers:
            headers = None
        file_response = self._session.post(
            url,
            expect_json_response=False,
            data=data,
            files=files,
            headers=headers,
        ).json()
        if 'entries' in file_response:
            file_response = file_response['entries'][0]
        return self.translator.translate(
            session=self._session,
            response_object=file_response,
        )

    @api_call
    def update_contents(
            self,
            file_path: str,
            etag: Optional[str] = None,
            preflight_check: bool = False,
            preflight_expected_size: int = 0,
            upload_using_accelerator: bool = False,
            file_name: Optional[str] = None,
            content_modified_at: Union[datetime, str] = None,
            additional_attributes: Optional[dict] = None,
            sha1: Optional[str] = None,
    ) -> 'File':
        """Upload a new version of a file. The contents are taken from the given file path.

        :param file_path:
            The path of the file that should be uploaded.
        :param etag:
            If specified, instruct the Box API to update the item only if the current version's etag matches.
        :param preflight_check:
            If specified, preflight check will be performed before actually uploading the file.
        :param preflight_expected_size:
            The size of the file to be uploaded in bytes, which is used for preflight check. The default value is '0',
            which means the file size is unknown.
        :param upload_using_accelerator:
            If specified, the upload will try to use Box Accelerator to speed up the uploads for big files.
            It will make an extra API call before the actual upload to get the Accelerator upload url, and then make
            a POST request to that url instead of the default Box upload url. It falls back to normal upload endpoint,
            if cannot get the Accelerator upload url.

            Please notice that this is a premium feature, which might not be available to your app.
        :param file_name:
            The new name to give the file on Box.
        :param content_modified_at:
            A datetime string in a format supported by the dateutil library or a datetime.datetime object,
            which specifies when the file content was last modified.
            If no timezone info provided, local timezone will be applied.
        :param additional_attributes:
            A dictionary containing attributes to add to the file that are not covered by other parameters.
        :param sha1:
            A sha1 checksum for the new content.
        :returns:
            A new file object
        :raises:
            :class:`BoxAPIException` if the specified etag doesn't match the latest version of the file or preflight
            check fails.
        """
        with open(file_path, 'rb') as file_stream:
            return self.update_contents_with_stream(
                file_stream,
                etag,
                preflight_check,
                preflight_expected_size=preflight_expected_size,
                upload_using_accelerator=upload_using_accelerator,
                file_name=file_name,
                content_modified_at=content_modified_at,
                additional_attributes=additional_attributes,
                sha1=sha1,
            )

    @api_call
    def lock(self, prevent_download: bool = False, expire_time: Union[datetime, str] = None) -> 'File':
        """
        Lock a file, preventing others from modifying (or possibly even downloading) it.

        :param prevent_download:
            Whether or not the lock should prevent other users from downloading the file.
        :param expire_time:
            A datetime string in a format supported by the dateutil library or a datetime.datetime object,
            which specifies when the lock should automatically expire, unlocking the file.
            If no timezone info provided, local timezone will be applied.
        :return:
            A new :class:`File` instance reflecting that the file has been locked.
        """
        data = {
            'lock': {
                'type': 'lock',
                'is_download_prevented': prevent_download,
            }
        }
        if expire_time is not None:
            data['lock']['expires_at'] = normalize_date_to_rfc3339_format(expire_time)
        return self.update_info(data=data)

    @api_call
    def unlock(self) -> 'File':
        """
        Unlock a file, releasing any restrictions that the lock maintained.

        :return:
            A new :class:`File` instance reflecting that the file has been unlocked.
        """
        data = {'lock': None}
        return self.update_info(data=data)

    @api_call
    def create_shared_link(
            self,
            *,
            access: Optional[str] = None,
            etag: Optional[str] = None,
            unshared_at: Union[datetime, str, None] = SDK_VALUE_NOT_SET,
            allow_download: Optional[bool] = None,
            allow_preview: Optional[bool] = None,
            allow_edit: Optional[bool] = None,
            password: Optional[str] = None,
            vanity_name: Optional[str] = None,
            **kwargs: Any
    ) -> 'File':
        """
        Baseclass override.

        :param access:
            Determines who can access the shared link. May be open, company, or collaborators. If no access is
            specified, the default access will be used.
        :param etag:
            If specified, instruct the Box API to create the link only if the current version's etag matches.
        :param unshared_at:
            The date on which this link should be disabled. May only be set if the current user is not a free user
            and has permission to set expiration dates. Takes a datetime string supported by the dateutil library
            or a datetime.datetime object. If no timezone info provided, local timezone will be applied.
            The time portion can be omitted, which defaults to midnight (00:00:00) on that date.
        :param allow_download:
            Whether the file being shared can be downloaded when accessed via the shared link.
            If this parameter is None, the default setting will be used.
        :param allow_preview:
            Whether the file being shared can be previewed when accessed via the shared link.
            If this parameter is None, the default setting will be used.
        :param allow_edit:
            Whether the file being shared can be edited when accessed via the shared link.
            If this parameter is None, the default setting will be used.
        :param password:
            The password required to view this link. If no password is specified then no password will be set.
            Please notice that this is a premium feature, which might not be available to your app.
        :param vanity_name:
            Defines a custom vanity name to use in the shared link URL, eg. https://app.box.com/v/my-custom-vanity-name.
            If this parameter is None, the standard shared link URL will be used.
        :param kwargs:
            Used to fulfill the contract of overriden method
        :return:
            The updated object with shared link.
            Returns a new object of the same type, without modifying the original object passed as self.
        :raises: :class:`BoxAPIException` if the specified etag doesn't match the latest version of the file.
        """
        # pylint:disable=arguments-differ
        return super().create_shared_link(
            access=access,
            etag=etag,
            unshared_at=unshared_at,
            allow_download=allow_download,
            allow_preview=allow_preview,
            allow_edit=allow_edit,
            password=password,
            vanity_name=vanity_name
        )

    @api_call
    def get_shared_link(
            self,
            *,
            access: Optional[str] = None,
            etag: Optional[str] = None,
            unshared_at: Union[datetime, str, None] = SDK_VALUE_NOT_SET,
            allow_download: Optional[bool] = None,
            allow_preview: Optional[bool] = None,
            allow_edit: Optional[bool] = None,
            password: Optional[str] = None,
            vanity_name: Optional[str] = None,
            **kwargs: Any
    ) -> 'str':
        """
        Baseclass override.

        :param access:
            Determines who can access the shared link. May be open, company, or collaborators. If no access is
            specified, the default access will be used.
        :param etag:
            If specified, instruct the Box API to create the link only if the current version's etag matches.
        :param unshared_at:
            The date on which this link should be disabled. May only be set if the current user is not a free user
            and has permission to set expiration dates. Takes a datetime string supported by the dateutil library
            or a datetime.datetime object. If no timezone info provided, local timezone will be applied.
            The time portion can be omitted, which defaults to midnight (00:00:00) on that date.
        :param allow_download:
            Whether the file being shared can be downloaded when accessed via the shared link.
            If this parameter is None, the default setting will be used.
        :param allow_preview:
            Whether the file being shared can be previewed when accessed via the shared link.
            If this parameter is None, the default setting will be used.
        :param allow_edit:
            Whether the file being shared can be edited when accessed via the shared link.
            If this parameter is None, the default setting will be used.
        :param password:
            The password required to view this link. If no password is specified then no password will be set.
            Please notice that this is a premium feature, which might not be available to your app.
        :param vanity_name:
            Defines a custom vanity name to use in the shared link URL, eg. https://app.box.com/v/my-custom-vanity-name.
            If this parameter is None, the standard shared link URL will be used.
        :param kwargs:
            Used to fulfill the contract of overriden method
        :returns:
            The URL of the shared link.
        :raises: :class:`BoxAPIException` if the specified etag doesn't match the latest version of the file.
        """
        # pylint:disable=arguments-differ
        return super().get_shared_link(
            access=access,
            etag=etag,
            unshared_at=unshared_at,
            allow_download=allow_download,
            allow_preview=allow_preview,
            allow_edit=allow_edit,
            password=password,
            vanity_name=vanity_name
        )

    @api_call
    def get_shared_link_download_url(
            self,
            access: Optional[str] = None,
            etag: Optional[str] = None,
            unshared_at: Union[datetime, str, None] = SDK_VALUE_NOT_SET,
            allow_preview: Optional[bool] = None,
            password: Optional[str] = None,
            vanity_name: Optional[str] = None
    ) -> str:
        """
        Get a shared link download url for the file with the given access permissions.
        This url is a direct download url for the file.

        :param access:
            Determines who can access the shared link. May be open, company, or collaborators. If no access is
            specified, the default access will be used.
        :param etag:
            If specified, instruct the Box API to create the link only if the current version's etag matches.
        :param unshared_at:
            The date on which this link should be disabled. May only be set if the current user is not a free user
            and has permission to set expiration dates. Takes a datetime string supported by the dateutil library
            or a datetime.datetime object. If no timezone info provided, local timezone will be applied.
            The time portion can be omitted, which defaults to midnight (00:00:00) on that date.
        :param allow_preview:
            Whether or not the item being shared can be previewed when accessed via the shared link.
            If this parameter is None, the default setting will be used.
        :param password:
            The password required to view this link. If no password is specified then no password will be set.
            Please notice that this is a premium feature, which might not be available to your app.
        :param vanity_name:
            Defines a custom vanity name to use in the shared link URL, eg. https://app.box.com/v/my-custom-vanity-name.
            If this parameter is None, the standard shared link URL will be used.
        :returns:
            The URL of the shared link that allows direct download.
        :raises: :class:`BoxAPIException` if the specified etag doesn't match the latest version of the item.
        """
        item = self.create_shared_link(
            access=access,
            etag=etag,
            unshared_at=unshared_at,
            allow_preview=allow_preview,
            password=password,
            vanity_name=vanity_name
        )
        return item.shared_link['download_url']  # pylint:disable=no-member

    @api_call
    def get_comments(
            self,
            limit: Optional[int] = None,
            offset: int = 0,
            fields: Iterable[str] = None
    ) -> 'BoxObjectCollection':
        """
        Get the comments on the file.

        :param limit:
            The maximum number of items to return per page. If not specified, then will use the server-side default.
        :param offset:
            The index at which to start returning items.
        :param fields:
            List of fields to request.
        :returns:
            An iterator of the items in the folder.
        """
        return LimitOffsetBasedObjectCollection(
            self.session,
            self.get_url('comments'),
            limit=limit,
            fields=fields,
            offset=offset,
            return_full_pages=False,
        )

    @api_call
    def add_comment(self, message: str) -> 'Comment':
        """
        Add a comment to the file.

        :param message:
            The content of the reply comment.
        :return: Added comment
        """
        url = self._session.get_url('comments')
        comment_class = self._session.translator.get('comment')
        data = comment_class.construct_params_from_message(message)
        data['item'] = {
            'type': 'file',
            'id': self.object_id
        }
        box_response = self._session.post(url, data=json.dumps(data))
        response = box_response.json()
        return self._session.translator.translate(
            session=self._session,
            response_object=response,
        )

    @api_call
    def create_task(
            self,
            message: Optional[str] = None,
            due_at: Union[datetime, str] = None,
            action: str = 'review',
            completion_rule: Optional[str] = None
    ) -> 'Task':
        """
        Create a task on the given file.

        :param message:
            An optional message to include in the task.
        :param due_at:
            When this task is due. Takes a datetime string supported by the dateutil library
            or a datetime.datetime object. If no timezone info provided, local timezone will be applied
        :param action:
            The type of task the task assignee will be prompted to perform.
            Value is one of review,complete
        :param completion_rule:
            Defines which assignees need to complete this task before the task
            is considered completed.
            Value is one of all_assignees,any_assignee
        :return:
            The newly created task
        """
        url = self._session.get_url('tasks')
        task_attributes = {
            'item': {
                'type': 'file',
                'id': self.object_id
            },
            'action': action,
        }
        if message is not None:
            task_attributes['message'] = message
        if due_at is not None:
            task_attributes['due_at'] = normalize_date_to_rfc3339_format(due_at)
        if completion_rule is not None:
            task_attributes['completion_rule'] = completion_rule
        box_response = self._session.post(url, data=json.dumps(task_attributes))
        response = box_response.json()
        return self.translator.translate(
            session=self._session,
            response_object=response,
        )

    @api_call
    def get_tasks(self, fields: Iterable[str] = None) -> 'BoxObjectCollection':
        """
        Get the entries in the file tasks.

        :param fields:
            List of fields to request.
        :returns:
            An iterator of the entries in the file tasks
        """
        return MarkerBasedObjectCollection(
            session=self._session,
            url=self.get_url('tasks'),
            limit=None,
            marker=None,
            fields=fields,
            return_full_pages=False,
        )

    @api_call
    def get_previous_versions(
            self,
            limit: Optional[int] = None,
            offset: int = None,
            fields: Iterable[str] = None
    ) -> 'BoxObjectCollection':
        """
        Get previous versions of the file.

        :param limit:
            The maximum number of items to return per page. If not specified, then will use the server-side default.
        :param offset:
            The index at which to start returning items.
        :param fields:
            List of fields to request.
        :returns:
            An iterator of the previous versions of the file.
        """
        return LimitOffsetBasedObjectCollection(
            session=self.session,
            url=self.get_url('versions'),
            limit=limit,
            fields=fields,
            offset=offset,
            return_full_pages=False,
        )

    @api_call
    def promote_version(self, file_version: 'FileVersion') -> 'FileVersion':
        """
        Promote a file version to become the current version of this file.  This will create a new file version
        identical to the previous version as the new current version.

        :param file_version:
            The file version to promote.
        :returns:
            The new file version created as the current.
        """
        url = self.get_url('versions', 'current')
        body = {
            'type': 'file_version',
            'id': file_version.object_id,
        }
        response = self._session.post(url, data=json.dumps(body)).json()
        return self.translator.translate(
            session=self._session,
            response_object=response,
        )

    @api_call
    def delete_version(self, file_version: 'FileVersion', etag: Optional[str] = None) -> bool:
        """
        Delete a specific version of a file.

        :param file_version:
            The file version to delete.
        :param etag:
            If specified, instruct the Box API to update the item only if the current version's etag matches.
        :returns:
            Whether the operation succeeded.
        """
        url = self.get_url('versions', file_version.object_id)
        headers = {'If-Match': etag} if etag is not None else None
        response = self._session.delete(url, expect_json_response=False, headers=headers)
        return response.ok

    @api_call
    def get_embed_url(self) -> str:
        """
        Get a URL suitable for embedding the file in an iframe in a web application.

        :returns:
            The embed URL.
        """
        url = self.get_url()
        params = {'fields': 'expiring_embed_link'}
        response = self._session.get(url, params=params).json()
        return response['expiring_embed_link']['url']

    @api_call
    def get_representation_info(self, rep_hints: Optional[str] = None) -> List[dict]:
        """
        Get information about the representations available for a file.

        :param rep_hints:
            A formatted string describing which representations are desired.
        :returns:
            The representation information
        """
        url = self.get_url()
        params = {'fields': 'representations'}
        headers = {'X-Rep-Hints': rep_hints} if rep_hints is not None else None
        response = self._session.get(url, params=params, headers=headers).json()
        return response['representations']['entries']

    @deprecated('Use get_thumbnail_representation')
    @api_call
    def get_thumbnail(
            self,
            extension: str = 'png',
            min_width: Optional[int] = None,
            min_height: Optional[int] = None,
            max_width: Optional[int] = None,
            max_height: Optional[int] = None
    ) -> bytes:
        """
        Retrieve a thumbnail image for the file.

        :param extension:
            The file extension for the thumbnail, e.g. 'png' or 'jpg'
        :param min_width:
            The minimum width required for the thumbnail image
        :param min_height:
            The minimum height required for the thumbnail image
        :param max_width:
            The maximum width required for the thumbnail image
        :param max_height:
            The maximum height required for the thumbnail image
        :returns:
            The file contents of the thumbnail image
        """
        url = self.get_url('thumbnail.' + extension)
        params = {}
        if min_width is not None:
            params['min_width'] = min_width
        if min_height is not None:
            params['min_height'] = min_height
        if max_width is not None:
            params['max_width'] = max_width
        if max_height is not None:
            params['max_height'] = max_height

        response = self._session.get(url, params=params, expect_json_response=False)
        return response.content

    @api_call
    def get_thumbnail_representation(self, dimensions: str, extension: str = 'png') -> bytes:
        """
        Retrieve a thumbnail image for the file.

        :param dimensions:
            The width by height size of this representation in pixels (e.g. '92x92')
        :param extension:
            The file extension for the thumbnail, e.g. 'png' or 'jpg'
        :returns:
            The file contents of the thumbnail image
        """
        rep_hints = f'[{extension}?dimensions={dimensions}]'
        representations = self.get_representation_info(rep_hints)
        if representations:
            representation = representations[0]
            if representation['status'].get('code') in ('error_conversion_failed', 'error_password_protected'):
                return b''
            url = representation['content']['url_template']
            url = url.replace('{+asset_path}', '')
            response = self._session.get(url, expect_json_response=False)
            return response.content
        return b''

    @api_call
    def copy(
            self,
            *,
            parent_folder: 'Folder',
            name: Optional[str] = None,
            file_version: 'FileVersion' = None,
            **_kwargs
    ) -> 'File':
        # pylint: disable=arguments-differ
        """Copy the item to the given folder.

        :param parent_folder:
            The folder to which the item should be copied.
        :param name:
            A new name for the item, in case there is already another item in the new parent folder with the same name.
        :param file_version:
            A specific version of the file to copy
        :returns:
            The copy of the file
        """
        # pylint: disable=arguments-differ
        url = self.get_url('copy')
        data = {
            'parent': {'id': parent_folder.object_id}
        }
        if name is not None:
            data['name'] = name
        if file_version is not None:
            data['version'] = file_version.object_id
        box_response = self._session.post(url, data=json.dumps(data))
        response = box_response.json()
        return self.translator.translate(
            session=self._session,
            response_object=response,
        )

    @api_call
    def set_disposition_at(self, date_time: Union[datetime, str]) -> 'File':
        """
        Modifies the retention expiration timestamp for the given file. This date can't be shortened once set on a file.

        :param date_time:
            A datetime string in a format supported by the dateutil library or a datetime.datetime object.
            If no timezone info provided, local timezone will be applied.
        :return:
            Updated 'File' object
        """
        data = {'disposition_at': normalize_date_to_rfc3339_format(date_time)}
        return self.update_info(data=data)



from functools import wraps
import json
import jsonpatch
from time import sleep

from bottle import template

from test.functional.mock_box.util.http_utils import abort


def allow_chaos(method):
    """Decorator for a method to allow erroneous operation."""
    method.call_number = 0
    added_chaos = []

    def add_chaos(chaos, should_apply=lambda call_number, *_: True):
        added_chaos.append((chaos, should_apply))

    def reset_chaos(call_number=0):
        del added_chaos[:]
        method.call_number = call_number

    @wraps(method)
    def chaotic_method(self, *args, **kwargs):
        skip_chaos = kwargs.pop('skip_chaos', False)
        method_to_apply = method
        if not skip_chaos:
            method.call_number += 1
            for chaos, should_apply in added_chaos:
                should_apply_result = False
                if isinstance(should_apply, int):
                    should_apply_result = method.call_number == should_apply
                elif hasattr(should_apply, '__iter__'):
                    should_apply_result = method.call_number in should_apply
                elif hasattr(should_apply, '__call__'):
                    should_apply_result = should_apply(method.call_number, args, kwargs)
                if should_apply_result:
                    method_to_apply = chaos(method_to_apply)
        return method_to_apply(self, *args, **kwargs)

    chaotic_method.add_chaos = add_chaos
    chaotic_method.reset_chaos = reset_chaos

    return chaotic_method


def delay(delay_seconds):
    return lambda delayed_function: lambda *args, **kwargs: sleep(delay_seconds) or delayed_function(*args, **kwargs)


def error(code, message=None, headers=None):
    return lambda erroneous_function: lambda *args, **kwargs: abort(code, message=message, headers=headers)


def html(method):
    return lambda *args, **kwargs: template('html_response', response=method(*args, **kwargs))


def xml(method):
    return lambda *args, **kwargs: template('xml_response', response=method(*args, **kwargs))


def patch(operations):

    json_patch = jsonpatch.JsonPatch(operations)

    def patcher(doc):
        return json_patch.apply(doc)

    def inner(patched_function):
        def patched_inner(*args, **kwargs):
            return_value = patched_function(*args, **kwargs)
            not_json = False
            if not isinstance(return_value, str):
                return_value = json.dumps(return_value)
                not_json = True
            return_value = patcher(return_value)
            if not_json:
                return_value = json.loads(return_value)
            return return_value
        return patched_inner
    return inner



from datetime import datetime
import pytest

from boxsdk import BoxAPIException
from test.integration_new.context_managers.box_retention_policy import BoxRetentionPolicy
from test.integration_new.context_managers.box_test_file import BoxTestFile
from test.integration_new.context_managers.box_test_folder import BoxTestFolder
from test.integration_new.context_managers.box_metadata_template import BoxTestMetadataTemplate


RETENTION_POLICY_ASSIGNMENT_TESTS_DIRECTORY_NAME = 'retention-policy-assignment-integration-tests'


@pytest.fixture(scope="module", autouse=True)
def parent_folder():
    with BoxTestFolder(name=f'{RETENTION_POLICY_ASSIGNMENT_TESTS_DIRECTORY_NAME} {datetime.now()}') as folder:
        yield folder


@pytest.fixture(scope="module", autouse=True)
def test_file(parent_folder, small_file_path):
    with BoxTestFile(parent_folder=parent_folder, file_path=small_file_path) as file:
        yield file


def test_delete_retention_policy_assignment(parent_folder, small_file_path):
    with BoxRetentionPolicy(disposition_action='permanently_delete', retention_length=1) as retention_policy:
        with BoxTestFolder(name=f'{RETENTION_POLICY_ASSIGNMENT_TESTS_DIRECTORY_NAME} {datetime.now()}') as folder_under_retention:
            retention_policy_assignment = retention_policy.assign(folder_under_retention)

            assignment = retention_policy_assignment.get()
            assert assignment.id is not None

            retention_policy_assignment.delete()

            with pytest.raises(BoxAPIException):
                retention_policy_assignment.get()


def test_retention_policy_assignement_to_metadata_template():
    with BoxTestMetadataTemplate(display_name="test_template") as metadata_template:
        with BoxRetentionPolicy(disposition_action='permanently_delete', retention_length=1) as retention_policy:
            retention_policy_assignment = retention_policy.assign(metadata_template, start_date_field='upload_date')

            assignment = retention_policy_assignment.get()
            assert assignment.id is not None
            assert assignment.start_date_field == 'upload_date'

            retention_policy_assignment.delete()

            with pytest.raises(BoxAPIException):
                retention_policy_assignment.get()



from unittest.mock import Mock
import pytest

from boxsdk import Client
from boxsdk.auth.oauth2 import OAuth2
from boxsdk.session.session import Session, AuthorizedSession

from .mock_network import MockNetwork


@pytest.fixture()
def box_client(box_oauth, mock_box_session):
    # pylint:disable=redefined-outer-name
    return Client(box_oauth, session=mock_box_session)


@pytest.fixture()
def box_oauth(unauthorized_session, client_id, client_secret, access_token, refresh_token):
    # pylint:disable=redefined-outer-name
    return OAuth2(
        client_id,
        client_secret,
        session=unauthorized_session,
        access_token=access_token,
        refresh_token=refresh_token,
    )


@pytest.fixture()
def mock_box_network():
    return MockNetwork()


@pytest.fixture()
def unauthorized_session(mock_box_network):
    return Session(network_layer=mock_box_network)


@pytest.fixture()
def mock_box_session(mock_box_network, box_oauth):
    return AuthorizedSession(box_oauth, network_layer=mock_box_network)


@pytest.fixture
def generic_successful_response(generic_successful_request_response):
    generic_successful_request_response.request = Mock()
    return generic_successful_request_response


@pytest.fixture
def successful_token_mock(successful_token_request_response):
    successful_token_request_response.request = Mock()
    return successful_token_request_response



from boxsdk.config import API


def test_get(mock_box_session, test_collaboration_allowlist_entry):
    entry_id = test_collaboration_allowlist_entry.object_id
    expected_url = f'{API.BASE_API_URL}/collaboration_whitelist_entries/{entry_id}'
    mock_entry = {
        'type': 'collaboration_whitelist_entry',
        'id': '98765',
        'domain': 'example.com',
        'direction': 'inbound'
    }
    mock_box_session.get.return_value.json.return_value = mock_entry
    entry = test_collaboration_allowlist_entry.get()
    mock_box_session.get.assert_called_once_with(expected_url, headers=None, params=None)
    assert entry.id == mock_entry['id']
    assert entry.domain == mock_entry['domain']
    assert entry.direction == mock_entry['direction']


def test_delete(mock_box_session, test_collaboration_allowlist_entry):
    entry_id = test_collaboration_allowlist_entry.object_id
    expected_url = f'{API.BASE_API_URL}/collaboration_whitelist_entries/{entry_id}'
    test_collaboration_allowlist_entry.delete()
    mock_box_session.delete.assert_called_once_with(expected_url, expect_json_response=False, headers=None, params={})



from unittest.mock import patch
import base64
import hashlib
import io
import json
import pytest

from boxsdk.exception import BoxAPIException
from boxsdk.config import API
from boxsdk.util.chunked_uploader import ChunkedUploader
from boxsdk.object.file import File
from boxsdk.object.upload_session import UploadSession


@pytest.fixture()
def test_upload_session(mock_box_session):
    upload_session_response_object = {
        'part_size': 8,
        'total_parts': 10,
    }
    return UploadSession(mock_box_session, '11493C07ED3EABB6E59874D3A1EF3581', upload_session_response_object)


def test_get_parts(test_upload_session, mock_box_session):
    expected_url = f'{API.UPLOAD_URL}/files/upload_sessions/{test_upload_session.object_id}/parts'
    mock_entry = {
        'part_id': '8F0966B1',
        'offset': 0,
        'size': 8,
        'sha1': None,
    }
    mock_box_session.get.return_value.json.return_value = {
        'entries': [mock_entry],
        'offset': 0,
        'total_count': 1,
        'limit': 1000,
    }
    test_parts = test_upload_session.get_parts()
    part = test_parts.next()
    mock_box_session.get.assert_called_once_with(expected_url, params={'offset': None})
    assert isinstance(part, dict)
    assert part['part_id'] == mock_entry['part_id']
    assert part['size'] == mock_entry['size']
    assert part['offset'] == mock_entry['offset']


def test_abort(test_upload_session, mock_box_session):
    expected_url = f'{API.UPLOAD_URL}/files/upload_sessions/{test_upload_session.object_id}'
    mock_box_session.delete.return_value.ok = True
    result = test_upload_session.abort()
    mock_box_session.delete.assert_called_once_with(expected_url, expect_json_response=False, headers=None, params={})
    assert result is True


def test_upload_part_bytes(test_upload_session, mock_box_session):
    expected_url = f'{API.UPLOAD_URL}/files/upload_sessions/{test_upload_session.object_id}'
    part_bytes = b'abcdefgh'
    offset = 32
    total_size = 80
    expected_sha1 = 'QlrxKgdDUCsyLpOgFbz4aOMk1Wo='
    expected_headers = {
        'Content-Type': 'application/octet-stream',
        'Digest': f'SHA={expected_sha1}',
        'Content-Range': 'bytes 32-39/80',
    }
    mock_box_session.put.return_value.json.return_value = {
        'part': {
            'part_id': 'ABCDEF123',
            'offset': offset,
            'size': 8,
            'sha1': expected_sha1,
        },
    }
    part = test_upload_session.upload_part_bytes(part_bytes, offset, total_size)

    mock_box_session.put.assert_called_once_with(expected_url, data=part_bytes, headers=expected_headers)
    assert isinstance(part, dict)
    assert part['sha1'] == expected_sha1
    assert part['size'] == 8
    assert part['part_id'] == 'ABCDEF123'


def test_commit(test_upload_session, mock_box_session):
    expected_url = f'{API.UPLOAD_URL}/files/upload_sessions/{test_upload_session.object_id}/commit'
    sha1 = hashlib.sha1()
    sha1.update(b'fake_file_data')
    file_id = '12345'
    file_type = 'file'
    file_etag = '7'
    file_attributes = {'description': 'This is a test description.'}
    parts = [
        {
            'part_id': 'ABCDEF123',
            'offset': 0,
            'size': 8,
            'sha1': 'fake_sha1',
        },
        {
            'part_id': 'ABCDEF456',
            'offset': 8,
            'size': 8,
            'sha1': 'fake_sha1',
        },
    ]
    expected_data = {
        'attributes': file_attributes,
        'parts': parts,
    }
    expected_headers = {
        'Content-Type': 'application/json',
        'Digest': f'SHA={base64.b64encode(sha1.digest()).decode("utf-8")}',
        'If-Match': '7',
    }
    mock_box_session.post.return_value.json.return_value = {
        'entries': [
            {
                'type': file_type,
                'id': file_id,
                'description': 'This is a test description.',
            },
        ],
    }
    created_file = test_upload_session.commit(content_sha1=sha1.digest(), parts=parts, file_attributes=file_attributes, etag=file_etag)
    mock_box_session.post.assert_called_once_with(expected_url, data=json.dumps(expected_data), headers=expected_headers)
    assert isinstance(created_file, File)
    assert created_file.id == file_id
    assert created_file.type == file_type
    assert created_file.description == 'This is a test description.'


def test_commit_with_missing_params(test_upload_session, mock_box_session):
    expected_get_url = f'{API.UPLOAD_URL}/files/upload_sessions/{test_upload_session.object_id}/parts'
    expected_url = f'{API.UPLOAD_URL}/files/upload_sessions/{test_upload_session.object_id}/commit'
    sha1 = hashlib.sha1()
    sha1.update(b'fake_file_data')
    file_id = '12345'
    file_type = 'file'
    parts = [
        {
            'part_id': '8F0966B1',
            'offset': 0,
            'size': 8,
            'sha1': None,
        },
    ]
    expected_data = {
        'parts': parts,
    }
    expected_headers = {
        'Content-Type': 'application/json',
        'Digest': f'SHA={base64.b64encode(sha1.digest()).decode("utf-8")}',
    }
    mock_entry = {
        'part_id': '8F0966B1',
        'offset': 0,
        'size': 8,
        'sha1': None,
    }
    mock_box_session.get.return_value.json.return_value = {
        'entries': [mock_entry],
        'offset': 0,
        'total_count': 1,
        'limit': 1000,
    }
    mock_box_session.post.return_value.json.return_value = {
        'entries': [
            {
                'type': file_type,
                'id': file_id,
            },
        ],
    }
    created_file = test_upload_session.commit(content_sha1=sha1.digest())
    mock_box_session.get.assert_called_once_with(expected_get_url, params={'offset': None})
    mock_box_session.post.assert_called_once_with(expected_url, data=json.dumps(expected_data), headers=expected_headers)
    assert isinstance(created_file, File)
    assert created_file.id == file_id
    assert created_file.type == file_type


def test_commit_returns_none_when_202_is_returned(test_upload_session, mock_box_session):
    expected_url = f'{API.UPLOAD_URL}/files/upload_sessions/{test_upload_session.object_id}/commit'
    sha1 = hashlib.sha1()
    sha1.update(b'fake_file_data')
    file_etag = '7'
    file_attributes = {'description': 'This is a test description.'}
    parts = [
        {
            'part_id': 'ABCDEF123',
            'offset': 0,
            'size': 8,
            'sha1': 'fake_sha1',
        },
        {
            'part_id': 'ABCDEF456',
            'offset': 8,
            'size': 8,
            'sha1': 'fake_sha1',
        },
    ]
    expected_data = {
        'attributes': file_attributes,
        'parts': parts,
    }
    expected_headers = {
        'Content-Type': 'application/json',
        'Digest': f'SHA={base64.b64encode(sha1.digest()).decode("utf-8")}',
        'If-Match': '7',
    }
    mock_box_session.post.side_effect = BoxAPIException(status=202)

    created_file = test_upload_session.commit(content_sha1=sha1.digest(), parts=parts, file_attributes=file_attributes, etag=file_etag)

    mock_box_session.post.assert_called_once_with(expected_url, data=json.dumps(expected_data), headers=expected_headers)
    assert created_file is None


def test_get_chunked_uploader_for_stream(test_upload_session):
    file_size = 197520
    part_bytes = b'abcdefgh'
    stream = io.BytesIO(part_bytes)
    chunked_uploader = test_upload_session.get_chunked_uploader_for_stream(stream, file_size)
    assert isinstance(chunked_uploader, ChunkedUploader)


def test_get_chunked_uploader(mock_content_response, mock_file_path, test_upload_session):
    mock_file_stream = io.BytesIO(mock_content_response.content)
    file_size = 197520
    with patch('os.stat') as stat:
        stat.return_value.st_size = file_size
        with patch('boxsdk.object.upload_session.open', return_value=mock_file_stream):
            chunked_uploader = test_upload_session.get_chunked_uploader(mock_file_path)
    assert isinstance(chunked_uploader, ChunkedUploader)



from datetime import datetime
from typing import Any, Optional, Iterable, Generator, Union, TYPE_CHECKING

from requests.exceptions import Timeout

from .base_endpoint import BaseEndpoint
from ..util.api_call_decorator import api_call
from ..util.datetime_formatter import normalize_date_to_rfc3339_format
from ..util.enum import ExtendableEnumMeta
from ..util.lru_cache import LRUCache
from ..util.text_enum import TextEnum

if TYPE_CHECKING:
    from boxsdk.object.event import Event
    from boxsdk.session.box_response import BoxResponse


# pylint:disable=too-many-ancestors
class EventsStreamType(TextEnum, metaclass=ExtendableEnumMeta):
    """An enum of all possible values of the `stream_type` parameter for user events.

    The value of the `stream_type` parameter determines the type of events
    returned by the endpoint.

    <https://developer.box.com/en/guides/events/>
    """


class UserEventsStreamType(EventsStreamType):
    """An enum of all possible values of the `stream_type` parameter for user events.

    - ALL: Returns all user events.
    - CHANGES: Returns tree changes.
    - SYNC: Returns tree changes only for sync folders.

    <https://developer.box.com/en/guides/events/for-user/>
    """
    ALL = 'all'
    CHANGES = 'changes'
    SYNC = 'sync'


class EnterpriseEventsStreamType(EventsStreamType):
    """An enum of all possible values of the `stream_type` parameter for enterprise events.

    - ADMIN_LOGS: Retrieves up to a year's events for all users in the enterprise. High latency.
    - ADMIN_LOGS_STREAMING: Retrieves up to a two weeks's events for all users in the enterprise. Low latency.

    NOTE: Requires Admin: These stream types will only work with an auth token
    from an enterprise admin account.

    <https://developer.box.com/en/guides/events/for-enterprise/>
    """
    ADMIN_LOGS = 'admin_logs'
    ADMIN_LOGS_STREAMING = 'admin_logs_streaming'
# pylint:enable=too-many-ancestors


class Events(BaseEndpoint):
    """Box API endpoint for subscribing to changes in a Box account."""

    def get_url(self, *args: Any) -> str:
        """Base class override."""
        return super().get_url('events', *args)

    @api_call
    def get_events(
            self,
            limit: int = 100,
            stream_position: Union[str, int] = 0,
            stream_type: EventsStreamType = UserEventsStreamType.ALL
    ) -> dict:
        """
        Get Box events from a given stream position for a given stream type.

        :param limit:
            Maximum number of events to return.
        :param stream_position:
            The location in the stream from which to start getting events. 0 is the beginning of time. 'now' will
            return no events and just current stream position.
        :param stream_type:
            (optional) Which type of events to return.
            Defaults to `UserEventsStreamType.ALL`.
        :returns:
            Dictionary containing the next stream position along with a list of some number of events.
        """
        url = self.get_url()
        params = {
            'limit': limit,
            'stream_position': stream_position,
            'stream_type': stream_type,
        }
        box_response = self._session.get(url, params=params)
        response = box_response.json().copy()
        return self.translator.translate(self._session, response_object=response)

    @api_call
    def get_admin_events(
            self,
            limit: Optional[int] = None,
            stream_position: Union[str, int] = 0,
            created_after: Union[datetime, str] = None,
            created_before: Union[datetime, str] = None,
            event_types: Iterable[str] = None
    ) -> dict:
        """
        Get Box Admin events from a datetime, to a datetime, or between datetimes with a given event type for a enterprise
        stream type. Used for historical querying (up to one year). Works for Enterprise admin_logs type.

        :param limit:
            (optional) Maximum number of events to return. If None, default API value limit=100 will be used.
        :param stream_position:
            The location in the stream from which to start getting events. 0 is the beginning of time.
        :param created_after:
            (optional) Start date in datetime format to pull events from
        :param created_before:
            (optional) End date in datetime format to pull events to
        :param event_types:
            (optional) Which events to return (ie. LOGIN)
        :returns:
            Dictionary containing the next stream position along with a list of some number of events.
        """
        url = self.get_url()
        params = {
            'created_after': normalize_date_to_rfc3339_format(created_after),
            'created_before': normalize_date_to_rfc3339_format(created_before),
            'stream_type': 'admin_logs',
        }
        if limit is not None:
            params['limit'] = limit
        if event_types is not None:
            params['event_type'] = ','.join(event_types)
        if stream_position is not None:
            params['stream_position'] = stream_position
        box_response = self._session.get(url, params=params)
        response = box_response.json()
        return self.translator.translate(self._session, response_object=response)

    @api_call
    def get_admin_events_streaming(
            self,
            limit: Optional[int] = None,
            stream_position: Union[str, int] = 0,
            event_types: Iterable[str] = None
    ) -> dict:
        """
        Get Box Admin events with a given event type for a enterprise stream type. Used for live monitoring (up to two weeks).
        Works for Enterprise admin_logs_streaming type.

        :param limit:
            (optional) Maximum number of events to return.
        :param stream_position:
            The location in the stream from which to start getting events. 0 is the beginning of time. 'now' will
            return no events and just current stream position.
        :param event_types:
            (optional) Which events to return (ie. LOGIN)
        :returns:
            Dictionary containing the next stream position along with a list of some number of events.
        """
        url = self.get_url()
        params = {
            'stream_type': 'admin_logs_streaming',
        }
        if limit is not None:
            params['limit'] = limit
        if stream_position is not None:
            params['stream_position'] = stream_position
        if event_types is not None:
            params['event_type'] = ','.join(event_types)
        box_response = self._session.get(url, params=params)
        response = box_response.json()
        return self.translator.translate(self._session, response_object=response)

    @api_call
    def get_latest_stream_position(self, stream_type: UserEventsStreamType = UserEventsStreamType.ALL) -> int:
        """
        Get the latest stream position. The return value can be used with :meth:`get_events` or
        :meth:`generate_events_with_long_polling`.

        :param stream_type:
            (optional) Which events stream to query.

            NOTE: Currently, the Box API requires this to be one of the user
            events stream types. The request will fail if an enterprise events
            stream type is passed.
        :returns:
            The latest stream position.
        """
        return self.get_events(limit=0, stream_position='now', stream_type=stream_type)['next_stream_position']

    def _get_all_events_since(
            self,
            stream_position: Union[str, int],
            stream_type: EventsStreamType = UserEventsStreamType.ALL
    ) -> Generator[tuple, None, None]:
        """
        :param stream_position:
            The location in the stream from which to start getting events. 0 is the beginning of time. 'now' will
            return no events and just current stream position.
        :param stream_type:
            (optional) Which type of events to return.
        """
        next_stream_position = stream_position
        while True:
            events = self.get_events(stream_position=next_stream_position, limit=100, stream_type=stream_type)
            next_stream_position = events['next_stream_position']
            events = events['entries']
            if not events:
                return
            for event in events:
                yield event, next_stream_position
            if len(events) < 100:
                return

    @api_call
    def long_poll(self, options: dict, stream_position: Union[str, int]) -> 'BoxResponse':
        """
        Set up a long poll connection at the specified url.

        :param options:
            The long poll options which include a long pull url, retry timeout, etc.
        :param stream_position:
            The location in the stream from which to start getting events. 0 is the beginning of time.
            'now' will return no events and just current stream position.
        :returns:
            {"message": "new_change"}, which means there're new changes on Box or {"version": 1, "message": "reconnect"}
            if nothing happens on Box during the long poll.
        """
        url = options['url']
        long_poll_response = self._session.get(
            url,
            timeout=options['retry_timeout'],
            params={'stream_position': stream_position}
        )
        return long_poll_response

    @api_call
    def generate_events_with_long_polling(
            self,
            stream_position: Union[str, int] = None,
            stream_type: UserEventsStreamType = UserEventsStreamType.ALL
    ) -> Generator['Event', None, None]:
        """
        Subscribe to events from the given stream position.

        :param stream_position:
            The location in the stream from which to start getting events. 0 is the beginning of time. 'now' will
            return no events and just current stream position.
        :param stream_type:
            (optional) Which type of events to return.

            NOTE: Currently, the Box API requires this to be one of the user
            events stream types. The request will fail if an enterprise events
            stream type is passed.
        :returns:
            Events corresponding to changes on Box in realtime, as they come in.
        """
        event_ids = LRUCache()
        stream_position = stream_position if stream_position is not None else self.get_latest_stream_position(stream_type=stream_type)
        while True:
            options = self.get_long_poll_options(stream_type=stream_type)
            while True:
                try:
                    long_poll_response = self.long_poll(options, stream_position)
                except Timeout:
                    break

                message = long_poll_response.json()['message']
                if message == 'new_change':
                    next_stream_position = stream_position
                    for event, next_stream_position in self._get_all_events_since(stream_position, stream_type=stream_type):
                        try:
                            event_ids.get(event['event_id'])
                        except KeyError:
                            yield event
                            event_ids.set(event['event_id'])
                    stream_position = next_stream_position
                    break
                if message == 'reconnect':
                    continue
                break

    @api_call
    def get_long_poll_options(self, stream_type: EventsStreamType = UserEventsStreamType.ALL) -> dict:
        """
        Get the url and retry timeout for setting up a long polling connection.

        :param stream_type:
            (optional) Which type of events to return.
        :returns:
            A `dict` including a long poll url, retry timeout, etc.
            E.g.
                {
                    "type": "realtime_server",
                    "url": "http://2.realtime.services.box.net/subscribe?channel=cc807c9c4869ffb1c81a&stream_type=all",
                    "ttl": "10",
                    "max_retries": "10",
                    "retry_timeout": 610,
                }
        """
        url = self.get_url()
        params = {'stream_type': stream_type}
        box_response = self._session.options(url, params=params)
        return box_response.json()['entries'][0]



import os
from datetime import datetime
from unittest.mock import Mock
import pytest
import pytz

from boxsdk.object.collaboration import Collaboration
from boxsdk.object.collection import Collection
from boxsdk.object.comment import Comment
from boxsdk.object.device_pinner import DevicePinner
from boxsdk.object.file import File
from boxsdk.object.file_request import FileRequest
from boxsdk.object.file_version import FileVersion
from boxsdk.object.file_version_retention import FileVersionRetention
from boxsdk.object.legal_hold import LegalHold
from boxsdk.object.folder import Folder
from boxsdk.object.folder_lock import FolderLock
from boxsdk.object.group import Group
from boxsdk.object.group_membership import GroupMembership
from boxsdk.object.legal_hold_policy import LegalHoldPolicy
from boxsdk.object.legal_hold_policy_assignment import LegalHoldPolicyAssignment
from boxsdk.object.metadata_template import MetadataTemplate
from boxsdk.object.user import User
from boxsdk.object.retention_policy import RetentionPolicy
from boxsdk.object.retention_policy_assignment import RetentionPolicyAssignment
from boxsdk.object.search import Search
from boxsdk.object.sign_request import SignRequest
from boxsdk.object.sign_template import SignTemplate
from boxsdk.object.storage_policy import StoragePolicy
from boxsdk.object.storage_policy_assignment import StoragePolicyAssignment
from boxsdk.object.terms_of_service import TermsOfService
from boxsdk.object.terms_of_service_user_status import TermsOfServiceUserStatus
from boxsdk.object.collaboration_allowlist import CollaborationAllowlist
from boxsdk.object.collaboration_allowlist_entry import CollaborationAllowlistEntry
from boxsdk.object.collaboration_allowlist_exempt_target import CollaborationAllowlistExemptTarget
from boxsdk.object.webhook import Webhook
from boxsdk.object.task import Task
from boxsdk.object.task_assignment import TaskAssignment
from boxsdk.object.web_link import WebLink
from boxsdk.util.default_arg_value import SDK_VALUE_NOT_SET

# pylint:disable=redefined-outer-name


@pytest.fixture(scope='module')
def mock_group_membership_id():
    return 'fake-group-membership-5'


@pytest.fixture(scope='module')
def mock_collaboration_id():
    return 'collab_id1'


@pytest.fixture(scope='module')
def mock_collection_id():
    return 'collection_id1'


@pytest.fixture(scope='module')
def mock_file_path():
    return os.path.join('path', 'to', 'file')


@pytest.fixture(scope='function')
def mock_image_path(image_extension):
    return os.path.join('path', 'to', f'image.{image_extension}')


@pytest.fixture(scope='function')
def mock_content_response(make_mock_box_request):
    mock_box_response, mock_network_response = make_mock_box_request(content=b'Contents of a text file.')
    mock_network_response.response_as_stream = raw = Mock()
    raw.stream.return_value = (bytes((b,)) for b in mock_box_response.content)
    return mock_box_response


@pytest.fixture(scope='function', params=[False, True])
def mock_upload_response_contains_entries(request):
    """Is the upload response formatted as {"type": "file", "id": "123", ...}, or as {"entries": [{...}]}.

    The v2.0 API does the latter, but future versions might do the former. So
    we'll test both.
    """
    return request.param


@pytest.fixture(scope='function')
def mock_upload_response(mock_object_id, make_mock_box_request, mock_upload_response_contains_entries):
    response = {'type': 'file', 'id': mock_object_id, 'description': 'Test File Description', }
    if mock_upload_response_contains_entries:
        response = {'entries': [response]}
    mock_box_response, _ = make_mock_box_request(response=response)
    return mock_box_response


@pytest.fixture()
def test_collaboration(mock_box_session, mock_collaboration_id):
    return Collaboration(mock_box_session, mock_collaboration_id)


@pytest.fixture()
def test_file(mock_box_session, mock_object_id):
    return File(mock_box_session, mock_object_id)


@pytest.fixture()
def test_file_version(mock_box_session):
    return FileVersion(mock_box_session, 'file_version_id')


@pytest.fixture()
def test_comment(mock_box_session, mock_object_id):
    return Comment(mock_box_session, mock_object_id)


@pytest.fixture()
def test_collaboration_allowlist(mock_box_session):
    return CollaborationAllowlist(mock_box_session)


@pytest.fixture()
def test_collaboration_allowlist_entry(mock_box_session, mock_object_id):
    return CollaborationAllowlistEntry(mock_box_session, mock_object_id)


@pytest.fixture()
def test_collaboration_allowlist_exemption(mock_box_session, mock_object_id):
    return CollaborationAllowlistExemptTarget(mock_box_session, mock_object_id)


@pytest.fixture()
def test_folder(mock_box_session, mock_object_id):
    return Folder(mock_box_session, mock_object_id)


@pytest.fixture()
def test_group(mock_box_session, mock_group_id):
    return Group(mock_box_session, mock_group_id)


@pytest.fixture()
def test_retention_policy(mock_box_session, mock_object_id):
    return RetentionPolicy(mock_box_session, mock_object_id)


@pytest.fixture()
def test_file_version_retention(mock_box_session, mock_object_id):
    return FileVersionRetention(mock_box_session, mock_object_id)


@pytest.fixture()
def test_retention_policy_assignment(mock_box_session, mock_object_id):
    return RetentionPolicyAssignment(mock_box_session, mock_object_id)


@pytest.fixture()
def test_group_membership(mock_box_session, mock_object_id):
    return GroupMembership(mock_box_session, mock_object_id)


@pytest.fixture()
def test_legal_hold_policy(mock_box_session, mock_object_id):
    return LegalHoldPolicy(mock_box_session, mock_object_id)


@pytest.fixture()
def test_legal_hold_policy_assignment(mock_box_session, mock_object_id):
    return LegalHoldPolicyAssignment(mock_box_session, mock_object_id)


@pytest.fixture()
def test_legal_hold(mock_box_session, mock_object_id):
    return LegalHold(mock_box_session, mock_object_id)


@pytest.fixture()
def test_search(mock_box_session):
    return Search(mock_box_session)


@pytest.fixture()
def test_storage_policy(mock_box_session, mock_object_id):
    return StoragePolicy(mock_box_session, mock_object_id)


@pytest.fixture()
def test_storage_policy_assignment(mock_box_session, mock_object_id):
    return StoragePolicyAssignment(mock_box_session, mock_object_id)


@pytest.fixture()
def test_terms_of_service(mock_box_session, mock_object_id):
    return TermsOfService(mock_box_session, mock_object_id)


@pytest.fixture()
def test_terms_of_service_user_status(mock_box_session, mock_object_id):
    return TermsOfServiceUserStatus(mock_box_session, mock_object_id)


@pytest.fixture()
def test_webhook(mock_box_session, mock_object_id):
    return Webhook(mock_box_session, mock_object_id)


@pytest.fixture()
def test_task(mock_box_session, mock_object_id):
    return Task(mock_box_session, mock_object_id)


@pytest.fixture()
def test_task_assignment(mock_box_session, mock_object_id):
    return TaskAssignment(mock_box_session, mock_object_id)


@pytest.fixture()
def test_web_link(mock_box_session, mock_object_id):
    return WebLink(mock_box_session, mock_object_id)


@pytest.fixture()
def test_device_pin(mock_box_session, mock_object_id):
    return DevicePinner(mock_box_session, mock_object_id)


@pytest.fixture(scope='function')
def mock_collab_response(make_mock_box_request, mock_collaboration_id):
    mock_box_response, _ = make_mock_box_request(
        response={'type': 'collaboration', 'id': mock_collaboration_id},
    )
    return mock_box_response


@pytest.fixture()
def mock_user(mock_box_session, mock_user_id):
    user = User(mock_box_session, mock_user_id)
    return user


@pytest.fixture()
def mock_collection(mock_box_session, mock_collection_id):
    return Collection(mock_box_session, mock_collection_id)


@pytest.fixture(scope='function')
def mock_precondition_failed_response(mock_object_id, make_mock_box_request):
    mock_box_response, _ = make_mock_box_request(
        status_code=412,
        response_ok=False,
        response={'type': 'folder', 'id': mock_object_id},
    )
    return mock_box_response


@pytest.fixture()
def mock_group_membership_dict(mock_group_membership_id, mock_user_id, mock_group_id):
    data = {
        'type': 'group_membership',
        'id': mock_group_membership_id,
        'role': 'member',
        'user': {'type': 'user', 'id': mock_user_id},
        'group': {'type': 'group', 'id': mock_group_id},
    }
    return data


@pytest.fixture()
def mock_add_member_response(mock_group_membership_dict, make_mock_box_request):
    mock_box_response, _ = make_mock_box_request(
        status_code=200,
        response_ok=True,
        response=mock_group_membership_dict,
    )
    return mock_box_response


@pytest.fixture()
def if_match_header(etag):
    return {'If-Match': etag} if etag is not None else None


@pytest.fixture()
def if_none_match_header(etag):
    return {'If-None-Match': etag} if etag is not None else None


@pytest.fixture()
def if_match_sha1_header(etag, sha1):
    headers = {}
    if etag is not None:
        headers['If-Match'] = etag
    if sha1 is not None:
        headers['Content-MD5'] = sha1
    if not headers:
        headers = None
    return headers


@pytest.fixture(params=[None, 'etag'])
def etag(request):
    return request.param


@pytest.fixture(params=[None, 'cf23df2207d99a74fbe169e3eba035e633b65d94'])
def sha1(request):
    return request.param


@pytest.fixture(params=[True, False])
def preflight_check(request):
    return request.param


@pytest.fixture(params=[True, False])
def upload_using_accelerator(request):
    return request.param


@pytest.fixture(params=[True, False])
def preflight_fails(preflight_check, request):
    return preflight_check and request.param


@pytest.fixture(params=[True, False])
def upload_using_accelerator_fails(upload_using_accelerator, request):
    return upload_using_accelerator and request.param


@pytest.fixture(params=[0, 100])
def file_size(request):
    return request.param


@pytest.fixture()
def mock_group(mock_box_session, mock_group_id):
    group = Group(mock_box_session, mock_group_id)
    return group


@pytest.fixture(params=(True, False, None))
def shared_link_can_download(request):
    return request.param


@pytest.fixture(params=(True, False, None))
def shared_link_can_preview(request):
    return request.param


@pytest.fixture(params=(True, False, None))
def shared_link_can_edit(request):
    return request.param


@pytest.fixture(params=('open', None))
def shared_link_access(request):
    return request.param


@pytest.fixture(params=('hunter2', None))
def shared_link_password(request):
    return request.param


@pytest.fixture(
    params=(
        '2018-10-31',
        '2018-10-31T00:00:00+14:00',
        None,
        SDK_VALUE_NOT_SET,
        datetime(2018, 10, 31, 23, 59, 59, tzinfo=pytz.timezone('US/Alaska'))
    )
)
def shared_link_unshared_at(request):
    return request.param


@pytest.fixture(params=('my-custom-vanity-name', None))
def shared_link_vanity_name(request):
    return request.param


@pytest.fixture(params=[
    # Test case for plain message
    (
        'message',
        'Hello there!'
    ),

    # Test case for tagged message
    (
        'tagged_message',
        '@[22222:Test User] Hi!'
    )
])
def comment_params(request):
    return request.param


@pytest.fixture()
def test_metadata_template(mock_box_session):
    fake_response = {
        'type': 'metadata_template',
        'scope': 'enterprise',
        'templateKey': 'vContract',
    }
    return MetadataTemplate(mock_box_session, None, fake_response)


@pytest.fixture()
def test_folder_lock(mock_box_session, mock_object_id):
    return FolderLock(mock_box_session, mock_object_id)


@pytest.fixture()
def test_sign_request(mock_box_session, mock_object_id):
    return SignRequest(mock_box_session, mock_object_id)


@pytest.fixture()
def test_file_request(mock_box_session, mock_object_id):
    return FileRequest(mock_box_session, mock_object_id)


@pytest.fixture()
def test_sign_template(mock_box_session, mock_object_id):
    return SignTemplate(mock_box_session, mock_object_id)



from collections import OrderedDict

from requests_toolbelt.multipart.encoder import MultipartEncoder


class MultipartStream(MultipartEncoder):
    """
    Subclass of the requests_toolbelt's :class:`MultipartEncoder` that ensures that data
    is encoded before files. This allows a server to process information in the data before
    receiving the file bytes.
    """
    def __init__(self, data, files):
        fields = OrderedDict()
        for k in data:
            fields[k] = data[k]
        for k in files:
            fields[k] = files[k]
        super().__init__(fields)



from typing import Optional

import attr

from ..util.log import sanitize_dictionary


@attr.s(slots=True)
class BoxRequest:
    """Represents a Box API request.

    :param url:                     The URL being requested.
    :param method:                  The HTTP method to use for the request.
    :param headers:                 HTTP headers to include with the request.
    :param auto_session_renewal:    Whether or not the session can be automatically renewed if the request fails.
    :param expect_json_response:    Whether or not the API response must be JSON.
    """
    url: str = attr.ib()
    method: Optional[str] = attr.ib(default='GET')
    headers: Optional[dict] = attr.ib(default=attr.Factory(dict))
    auto_session_renewal: Optional[bool] = attr.ib(default=True)
    expect_json_response: Optional[bool] = attr.ib(default=True)
    access_token: Optional[str] = attr.ib(default=None)

    def __repr__(self) -> str:
        return f'<BoxRequest for {self.method} {self.url} with headers {sanitize_dictionary(self.headers)}'



from typing import Optional

from .base_object import BaseObject
from ..pagination.marker_based_object_collection import MarkerBasedObjectCollection
from ..util.api_call_decorator import api_call


class RetentionPolicyAssignment(BaseObject):
    """Represents a Box retention policy assignment."""
    _item_type = 'retention_policy_assignment'

    @api_call
    def get_files_under_retention(
            self,
            limit: Optional[int] = None,
            marker: Optional[str] = None
    ) -> 'MarkerBasedObjectCollection':
        """
        Retrieves all files under retention for a retention policy assignment

        :param limit: the limit of retrieved entries per page. Default 100.
        :param marker: the paging marker to start paging from.
        :return: An iterator of the entries with information about all files under retention.
        """
        return MarkerBasedObjectCollection(
            self.session,
            self.get_url('files_under_retention'),
            limit=limit,
            marker=marker
        )

    @api_call
    def get_file_versions_under_retention(
            self,
            limit: Optional[int] = None,
            marker: Optional[str] = None
    ) -> 'MarkerBasedObjectCollection':
        """
        Retrieves all file versions under retention for a retention policy assignment

        :param limit: the limit of retrieved entries per page. Default 100.
        :param marker: the paging marker to start paging from.
        :return: An iterator of the entries with information about all files uversions nder retention.
        """
        return MarkerBasedObjectCollection(
            self.session,
            self.get_url('file_versions_under_retention'),
            limit=limit,
            marker=marker
        )



from boxsdk.config import API
from boxsdk.object.file import File
from boxsdk.object.file_version import FileVersion
from boxsdk.object.retention_policy_assignment import RetentionPolicyAssignment


def test_get_assignment(test_retention_policy_assignment, mock_box_session):
    expected_url = f'{API.BASE_API_URL}/retention_policy_assignments/{test_retention_policy_assignment.object_id}'
    mock_box_session.get.return_value.json.return_value = {
        'type': test_retention_policy_assignment.object_type,
        'id': test_retention_policy_assignment.object_id,
    }
    retention_policy_assignment = test_retention_policy_assignment.get()
    mock_box_session.get.assert_called_once_with(expected_url, headers=None, params=None)
    assert isinstance(retention_policy_assignment, RetentionPolicyAssignment)
    assert retention_policy_assignment['type'] == test_retention_policy_assignment.object_type
    assert retention_policy_assignment['id'] == test_retention_policy_assignment.object_id


def test_get_files_under_retention(test_retention_policy_assignment, test_file, mock_box_session):
    # given:
    test_marker = 'test_marker'
    test_limit = 50
    target_url = f"{API.BASE_API_URL}/retention_policy_assignments/{test_retention_policy_assignment.object_id}/files_under_retention"

    mock_box_session.get.return_value.json.return_value = {
        'limit': test_limit,
        'entries': [test_file],
        'next_marker': test_marker,
    }

    # when:
    files_under_retention = test_retention_policy_assignment.get_files_under_retention(
        limit=test_limit,
        marker=test_marker
    )
    file_under_retention = files_under_retention.next()

    # then:
    params = {
        'limit': test_limit,
        'marker': test_marker,
    }
    mock_box_session.get.assert_called_once_with(target_url, params=params)
    assert isinstance(file_under_retention, File)
    assert file_under_retention.object_id == test_file.object_id


def test_get_file_versions_under_retention(test_retention_policy_assignment, test_file_version, mock_box_session):
    # given:
    test_marker = 'test_marker'
    test_limit = 50
    target_url = f"{API.BASE_API_URL}/retention_policy_assignments/{test_retention_policy_assignment.object_id}/file_versions_under_retention"

    mock_box_session.get.return_value.json.return_value = {
        'limit': test_limit,
        'entries': [test_file_version],
        'next_marker': test_marker,
    }

    # when:
    file_versions_under_retention = test_retention_policy_assignment.get_file_versions_under_retention(
        limit=test_limit,
        marker=test_marker
    )
    file_version_under_retention = file_versions_under_retention.next()

    # then:
    params = {
        'limit': test_limit,
        'marker': test_marker,
    }
    mock_box_session.get.assert_called_once_with(target_url, params=params)
    assert isinstance(file_version_under_retention, FileVersion)
    assert file_version_under_retention.object_id == test_file_version.object_id


def test_delete_assignment(test_retention_policy_assignment, mock_box_session):
    expected_url = f'{API.BASE_API_URL}/retention_policy_assignments/{test_retention_policy_assignment.object_id}'
    mock_box_session.delete.return_value.ok = True

    is_assignment_deleted = test_retention_policy_assignment.delete()
    mock_box_session.delete.assert_called_once_with(expected_url, expect_json_response=False, params={}, headers=None)

    assert is_assignment_deleted



from .developer_token_client import DeveloperTokenClient
from .logging_client import LoggingClient


class DevelopmentClient(LoggingClient, DeveloperTokenClient):
    """
    Client subclass that uses developer token auth and logs requests and responses.
    Great for use in development!
    """



import json
from datetime import datetime
import pytest
import pytz

from boxsdk.config import API
from boxsdk.object.folder import Folder
from boxsdk.object.file_request import FileRequest
from boxsdk.object.file_request import StatusState


def test_get(test_file_request, mock_box_session):
    expected_url = f'{API.BASE_API_URL}/file_requests/{test_file_request.object_id}'
    mock_box_session.get.return_value.json.return_value = {
        'type': test_file_request.object_type,
        'id': test_file_request.object_id,
        'title': 'File Request'
    }
    file_request = test_file_request.get()
    mock_box_session.get.assert_called_once_with(expected_url, headers=None, params=None)
    assert isinstance(file_request, FileRequest)
    assert file_request['type'] == file_request.object_type
    assert file_request['id'] == file_request.object_id
    assert file_request['title'] == 'File Request'


def test_update(test_file_request, mock_box_session):
    new_title = 'New File Request Title'
    new_status = StatusState.INACTIVE
    expected_url = f'{API.BASE_API_URL}/file_requests/{test_file_request.object_id}'
    mock_box_session.put.return_value.json.return_value = {
        'type': test_file_request.object_type,
        'id': test_file_request.object_id,
        'title': new_title,
        'status': new_status,
    }
    data = {
        'title': new_title,
        'status': StatusState.INACTIVE,
    }
    file_request = test_file_request.update_info(data=data)
    mock_box_session.put.assert_called_once_with(expected_url, data=json.dumps(data), headers=None, params=None)
    assert isinstance(file_request, FileRequest)
    assert file_request['type'] == test_file_request.object_type
    assert file_request['id'] == test_file_request.object_id
    assert file_request['title'] == new_title
    assert file_request['status'] == StatusState.INACTIVE


@pytest.mark.parametrize('expires_at', [
    '2019-07-01T22:02:24+14:00',
    datetime(2019, 7, 1, 22, 2, 24, tzinfo=pytz.timezone('US/Alaska'))
])
def test_copy(test_file_request, mock_box_session, expires_at):
    new_folder_id = '100'
    expected_url = f'{API.BASE_API_URL}/file_requests/{test_file_request.object_id}/copy'
    expected_expires_at = '2019-07-01T22:02:24+14:00'
    mock_box_session.post.return_value.json.return_value = {
        'type': test_file_request.object_type,
        'id': test_file_request.object_id,
        'title': 'File Request Copied',
        'folder': {
            'type': 'folder',
            'id': new_folder_id,
        },
        'expires_at': expected_expires_at,
    }
    new_title = 'File Request Copied'
    new_folder = Folder(mock_box_session, object_id=new_folder_id)
    file_request = test_file_request.copy(title=new_title, folder=new_folder, expires_at=expires_at)
    data = {
        'folder': {
            'id': new_folder_id,
            'type': 'folder',
        },
        'title': new_title,
        'expires_at': expected_expires_at,
    }
    mock_box_session.post.assert_called_once_with(expected_url, data=json.dumps(data))
    assert isinstance(file_request, FileRequest)
    assert file_request['type'] == test_file_request.object_type
    assert file_request['title'] == 'File Request Copied'
    assert file_request['folder']['id'] == '100'
    assert file_request['folder']['type'] == 'folder'


def test_delete(test_file_request, mock_box_session):
    expected_url = f'{API.BASE_API_URL}/file_requests/{test_file_request.object_id}'
    test_file_request.delete()
    mock_box_session.delete.assert_called_once_with(expected_url, expect_json_response=False, headers=None, params={})



from typing import Any

from boxsdk.object.user import User
from test.integration_new import util
from test.integration_new import CLIENT


class BoxTestUser:

    def __init__(self, *, name: str = None, login: str = None):
        if name is None:
            name = util.random_name()

        self._user: User = CLIENT.create_user(name=name, login=login)

    def __enter__(self) -> User:
        return self._user

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        self._user.delete(notify=False, force=True)



import json
import pytest

from boxsdk.config import API
from boxsdk.object.enterprise import Enterprise
from boxsdk.object.invite import Invite


@pytest.fixture()
def test_enterprise(mock_box_session):
    return Enterprise(
        session=mock_box_session,
        object_id='test_enterprise_id',
    )


def test_invite_user(test_enterprise, mock_box_session):
    # pylint:disable=redefined-outer-name
    expected_url = f'{API.BASE_API_URL}/invites'
    test_user_login = 'test@user.com'
    expected_body = json.dumps({
        'enterprise': {
            'id': test_enterprise.object_id,
        },
        'actionable_by': {
            'login': test_user_login,
        },
    })
    invite_json = {
        'type': 'invite',
        'id': '11111',
        'status': 'pending',
    }
    mock_box_session.post.return_value.json.return_value = invite_json
    new_invite = test_enterprise.invite_user(test_user_login)
    mock_box_session.post.assert_called_once_with(expected_url, data=expected_body)
    assert isinstance(new_invite, Invite)
    assert new_invite.object_id == invite_json['id']
    assert new_invite._session == mock_box_session  # pylint: disable=protected-access
    assert new_invite.status == invite_json['status']



import time
from abc import ABC, abstractmethod
from datetime import datetime
from typing import Optional, Tuple, TYPE_CHECKING, Union, Any

from boxsdk.auth.oauth2 import OAuth2
from boxsdk.exception import BoxOAuthException
from boxsdk.config import API
from boxsdk.object.user import User

if TYPE_CHECKING:
    from boxsdk.network.network_interface import NetworkResponse


class ServerAuth(ABC, OAuth2):
    USER_SUBJECT_TYPE = 'user'
    ENTERPRISE_SUBJECT_TYPE = 'enterprise'

    def __init__(
            self,
            client_id: str,
            client_secret: str,
            enterprise_id: Optional[str] = None,
            user: Optional[Union[str, 'User']] = None,
            **kwargs: Any
    ):
        super().__init__(client_id=client_id, client_secret=client_secret, **kwargs)
        self._enterprise_id = enterprise_id
        self._user_id = self._normalize_user_id(user)

    def _refresh(self, access_token: str) -> Tuple[str, None]:
        """
        Base class override.

        Instead of refreshing an access token using a refresh token, we just issue a new JWT request.
        """
        # pylint:disable=unused-argument
        if self._user_id is None:
            new_access_token = self.authenticate_instance()
        else:
            new_access_token = self.authenticate_user()
        return new_access_token, None

    def authenticate_user(self, user: Union[str, 'User'] = None) -> str:
        """
        Get an access token for a User.

        May be one of this application's created App User. Depending on the
        configured User Access Level, may also be any other App User or Managed
        User in the enterprise.

        <https://developer.box.com/en/guides/applications/>
        <https://developer.box.com/en/guides/authentication/select/>

        :param user:
            (optional) The user to authenticate, expressed as a Box User ID or
            as a :class:`User` instance.

            If not given, then the most recently provided user ID, if
            available, will be used.
        :raises:
            :exc:`ValueError` if no user ID was passed and the object is not
            currently configured with one.
        :return:
            The access token for the user.
        """
        sub = self._normalize_user_id(user) or self._user_id
        if not sub:
            raise ValueError("authenticate_user: Requires the user ID, but it was not provided.")
        self._user_id = sub
        return self._authenticate(sub, self.USER_SUBJECT_TYPE)

    authenticate_app_user = authenticate_user

    def authenticate_instance(self, enterprise: Optional[str] = None) -> str:
        """
        Get an access token for a Box Developer Edition enterprise.

        :param enterprise:
            The ID of the Box Developer Edition enterprise.

            Optional if the value was already given to `__init__`,
            otherwise required.
        :raises:
            :exc:`ValueError` if `None` was passed for the enterprise ID here
            and in `__init__`, or if the non-`None` value passed here does not
            match the non-`None` value passed to `__init__`.
        :return:
            The access token for the enterprise which can provision/deprovision app users.
        """
        enterprises = [enterprise, self._enterprise_id]
        if not any(enterprises):
            raise ValueError("authenticate_instance: Requires the enterprise ID, but it was not provided.")
        if all(enterprises) and (enterprise != self._enterprise_id):
            raise ValueError(
                f"authenticate_instance: Given enterprise ID {enterprise!r}, "
                f"but {self} already has ID {self._enterprise_id!r}"
            )
        if not self._enterprise_id:
            self._enterprise_id = enterprise
        self._user_id = None
        return self._authenticate(self._enterprise_id, self.ENTERPRISE_SUBJECT_TYPE)

    def _authenticate(self, subject_id: str, subject_type: str) -> str:
        """
        Authenticate with server type authentication (JWT or CCG).
        If authorization fails because the expiration time is out of sync with the Box servers,
        retry using the time returned in the error response.
        Pass an enterprise ID to get an enterprise token (which can be used to provision/deprovision users),
        or a user ID to get a user token.

        :param subject_id:
            The enterprise ID or user ID to auth.
        :param subject_type:
            Either 'enterprise' or 'user'
        :return:
            The access token for the enterprise or app user.
        """
        attempt_number = 0
        date = None
        while True:
            try:
                return self._fetch_access_token(subject_id, subject_type, date)
            except BoxOAuthException as ex:
                network_response = ex.network_response
                code = network_response.status_code  # pylint: disable=maybe-no-member
                box_datetime = self._get_date_header(network_response)

                if attempt_number >= API.MAX_RETRY_ATTEMPTS:
                    raise ex

                if code == 429 or code >= 500:
                    date = None
                elif box_datetime is not None and self._is_auth_error_retryable(network_response):
                    date = box_datetime
                else:
                    raise ex

                time_delay = self._session.get_retry_after_time(
                    attempt_number,
                    network_response.headers.get('Retry-After', None)
                )
                time.sleep(time_delay)
                attempt_number += 1
                self._logger.debug('Retrying authentication request')

    @abstractmethod
    def _fetch_access_token(self, subject_id: str, subject_type: str, now_time: Optional[datetime] = None) -> str:
        pass

    @staticmethod
    def _get_date_header(network_response: 'NetworkResponse') -> Optional[datetime]:
        """
        Get datetime object for Date header, if the Date header is available.

        :param network_response:
            The response from the Box API that should include a Date header.
        :return:
            The datetime parsed from the Date header, or None if the header is absent or if it couldn't be parsed.
        """
        box_date_header = network_response.headers.get('Date', None)
        if box_date_header is not None:
            try:
                return datetime.strptime(box_date_header, '%a, %d %b %Y %H:%M:%S %Z')
            except ValueError:
                pass
        return None

    @staticmethod
    def _is_auth_error_retryable(network_response: 'NetworkResponse') -> bool:
        """
        Determine whether the network response indicates that the authorization request was rejected because of
        the exp or jti claim and can be retried. Exp claim error can happen if the current system time is too
        different from the Box server time. If got an error: "A unique 'jti' value is required",
         we also retry auth in order to use new 'jti' claim.

        Returns True if the status code is 400, the error code is invalid_grant, and the error description indicates
        a problem with the exp or jti claim; False, otherwise.

        :param network_response:
            The response from the Box API that should include a Date header.
        """
        status_code = network_response.status_code
        try:
            json_response = network_response.json()
        except ValueError:
            return False
        error_code = json_response.get('error', '')
        error_description = json_response.get('error_description', '')
        return status_code == 400 and error_code == 'invalid_grant' \
            and ('exp' in error_description or 'jti' in error_description)

    @classmethod
    def _normalize_user_id(cls, user: Any) -> Optional[str]:
        """Get a Box user ID from a selection of supported param types.

        :param user:
            An object representing the user or user ID.

            Currently supported types are `unicode` (which represents the user
            ID) and :class:`User`.

            If `None`, returns `None`.
        :raises:  :exc:`TypeError` for unsupported types.
        """
        if user is None:
            return None
        if isinstance(user, User):
            return user.object_id
        if isinstance(user, str):
            return str(user)
        raise TypeError(f"Got unsupported type {user.__class__.__name__!r} for user.")



import pytest

from boxsdk.config import API
from boxsdk.object.sign_request import SignRequest


@pytest.fixture(scope='module')
def mock_sign_request_response():
    # pylint:disable=redefined-outer-name
    mock_sign_request = {
        'id': '42',
        'type': 'sign-request',
        'are_reminders_enabled': 'true',
        'are_text_signatures_enabled': 'true',
        'auto_expire_at': '2021-04-26T08:12:13.982Z',
        'days_valid': '2',
        'email_message': 'Hello! Please sign the document below',
        'email_subject': 'Sign Request from Acme',
        'external_id': '123',
        'is_document_preparation_needed': 'true',
        'redirect_url': 'https://www.box.com/accepted',
        'declined_redirect_url': 'https://www.box.com/declined',
        'parent_folder': {
            'id': '12345',
            'type': 'folder',
            'etag': '1',
            'name': 'Contracts',
            'sequence_id': '3'
        },
        'prefill_tags': [
            {
                'document_tag_id': '1234',
                'text_value': 'text',
                'checkbox_value': 'true',
                'date_value': '2021-04-26T08:12:13.982Z'
            }
        ],
        'prepare_url': 'https://prepareurl.com',
        'sign_files': {
            'files': [
                {
                    'id': '12345',
                    'etag': '1',
                    'type': 'file',
                    'sequence_id': '3',
                    'name': 'Contract.pdf',
                    'sha1': '85136C79CBF9FE36BB9D05D0639C70C265C18D37',
                    'file_version': {
                        'id': '12345',
                        'type': 'file_version',
                        'sha1': '134b65991ed521fcfe4724b7d814ab8ded5185dc'
                    }
                }
            ],
            'is_ready_for_download': 'true'
        },
        'signers': [
            {
                'email': 'example@gmail.com',
                'role': 'signer',
                'is_in_person': 'true',
                'order': '2',
                'embed_url_external_user_id': '1234',
                'has_viewed_document': 'true',
                'signer_decision': {
                    'type': 'signed',
                    'finalized_at': '2021-04-26T08:12:13.982Z'
                },
                'inputs': [
                    {
                        'document_tag_id': '1234',
                        'text_value': 'text',
                        'checkbox_value': 'true',
                        'date_value': '2021-04-26T08:12:13.982Z',
                        'type': 'text',
                        'page_index': '4'
                    }
                ],
                'embed_url': 'https://example.com'
            }
        ],
        'signing_log': {
            'id': '12345',
            'type': 'file',
            'etag': '1',
            'file_version': {
                'id': '12345',
                'type': 'file_version',
                'sha1': '134b65991ed521fcfe4724b7d814ab8ded5185dc'
            },
            'name': 'Contract.pdf',
            'sequence_id': '3',
            'sha1': '85136C79CBF9FE36BB9D05D0639C70C265C18D37'
        },
        'source_files': [
            {
                'id': '12345',
                'etag': '1',
                'type': 'file',
                'sequence_id': '3',
                'name': 'Contract.pdf',
                'sha1': '85136C79CBF9FE36BB9D05D0639C70C265C18D37',
                'file_version': {
                    'id': '12345',
                    'type': 'file_version',
                    'sha1': '134b65991ed521fcfe4724b7d814ab8ded5185dc'
                }
            }
        ],
        'status': 'cancelled',
        'template_id': '123075213-af2c8822-3ef2-4952-8557-52d69c2fe9cb'
    }
    return mock_sign_request


def test_get_sign_request(test_sign_request, mock_box_session, mock_sign_request_response):
    expected_url = f'{API.BASE_API_URL}/sign_requests/{test_sign_request.object_id}'
    mock_box_session.get.return_value.json.return_value = mock_sign_request_response

    sign_request = test_sign_request.get()

    mock_box_session.get.assert_called_once_with(
        expected_url, headers=None, params=None)
    assert isinstance(sign_request, SignRequest)
    assert sign_request['id'] == test_sign_request.object_id


def test_cancel_sign_request(test_sign_request, mock_box_session, mock_sign_request_response):
    expected_url = f'{API.BASE_API_URL}/sign_requests/{test_sign_request.object_id}/cancel'
    mock_box_session.post.return_value.json.return_value = mock_sign_request_response

    sign_request = test_sign_request.cancel()

    mock_box_session.post.assert_called_once_with(expected_url)
    assert isinstance(sign_request, SignRequest)
    assert sign_request['id'] == test_sign_request.object_id
    assert sign_request['status'] == mock_sign_request_response['status']


def test_resend_sign_request(test_sign_request, mock_box_session):
    expected_url = f'{API.BASE_API_URL}/sign_requests/{test_sign_request.object_id}/resend'

    test_sign_request.resend()

    mock_box_session.post.assert_called_once_with(
        expected_url, expect_json_response=False, skip_retry_codes={202})



from typing import Any, Iterable

from test.integration_new import CLIENT
from boxsdk.object.web_link import WebLink


class BoxTestSignRequest:

    def __init__(self, *, files: Iterable, signers: Iterable, parent_folder_id: str):
        self._sign_request = CLIENT.create_sign_request(files, signers, parent_folder_id)

    def __enter__(self) -> WebLink:
        return self._sign_request

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        self._sign_request.cancel()



from test.integration_new import util
from test.integration_new import CLIENT
from boxsdk.exception import BoxAPIException


def test_create_legal_hold_policy():
    policy_name = 'Test Legal Hold Policy ' + util.random_name()
    description = 'Test Legal Hold Policy Description'
    filter_started_at = '2021-12-12T10:53:43-08:00'
    filter_ended_at = '2022-12-18T10:53:43-08:00'

    legal_hold_policy = CLIENT.create_legal_hold_policy(
        policy_name=policy_name,
        description=description,
        filter_started_at=filter_started_at,
        filter_ended_at=filter_ended_at
    )

    try:
        assert legal_hold_policy.policy_name == policy_name
        assert legal_hold_policy.description == description
        assert legal_hold_policy.filter_started_at == filter_started_at
        assert legal_hold_policy.filter_ended_at == filter_ended_at

        new_policy_name = 'Test Legal Hold Policy ' + util.random_name()
        new_policy_description = 'Test Legal Hold Policy Description'
        legal_hold_policy.update_info(data={
            'policy_name': new_policy_name,
            'description': new_policy_description
        })

        legal_hold_policy = CLIENT.legal_hold_policy(policy_id=legal_hold_policy.object_id).get()
        assert legal_hold_policy.policy_name == new_policy_name
        assert legal_hold_policy.description == new_policy_description
    finally:
        legal_hold_policy.delete()



import json
from typing import TYPE_CHECKING

from .base_object import BaseObject
from ..util.api_call_decorator import api_call

if TYPE_CHECKING:
    from boxsdk.object.invite import Invite


class Enterprise(BaseObject):

    """Represents a single enterprise."""
    _item_type = 'enterprise'

    @api_call
    def invite_user(self, user_email: str) -> 'Invite':
        """
        Invites an existing user to an Enterprise. The user must already have a Box account.

        :param user_email:
            The login email address of the user that will receive the invitation.
        :returns:
            The invitation record for the user
        """
        url = self._session.get_url('invites')
        body = {
            'enterprise': {
                'id': self.object_id,
            },
            'actionable_by': {
                'login': user_email,
            },
        }
        response = self._session.post(url, data=json.dumps(body)).json()
        return self.translator.translate(
            session=self._session,
            response_object=response,
        )



from typing import Any, Union

from boxsdk.object.retention_policy import RetentionPolicy
from test.integration_new import USER_CLIENT


class BoxRetentionPolicy:

    DEFAULT_RETENTION_POLICY_NAME = "modifiable_retention_policy_for_integration_tests"

    def __init__(
            self,
            *,
            name: str = DEFAULT_RETENTION_POLICY_NAME,
            disposition_action: str = 'permanently_delete',
            retention_length: Union[float, int] = 1
    ):

        policy_type = 'finite' if retention_length < float('inf') else 'infinite'
        policies_found = list(USER_CLIENT.get_retention_policies(policy_name=name, policy_type=policy_type))

        if policies_found and policies_found[0].get().status == 'active':
            self._retention_policy = policies_found[0]
        else:
            self._retention_policy: RetentionPolicy = USER_CLIENT.create_retention_policy(
                policy_name=name,
                disposition_action=disposition_action,
                retention_length=retention_length,
                retention_type='modifiable'
            )

    def __enter__(self) -> RetentionPolicy:
        return self._retention_policy

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        pass



from contextlib import contextmanager
from datetime import datetime, timedelta
import io
from itertools import cycle, product
import json
import random
import string
from unittest.mock import Mock, mock_open, patch, sentinel, call

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey, generate_private_key as generate_rsa_private_key
from cryptography.hazmat.primitives import serialization
import pytest
import pytz

from boxsdk.auth.jwt_auth import JWTAuth
from boxsdk.exception import BoxOAuthException
from boxsdk.config import API
from boxsdk.object.user import User


@pytest.fixture(params=[16, 32, 128])
def jti_length(request):
    return request.param


@pytest.fixture(params=('RS256', 'RS512'))
def jwt_algorithm(request):
    return request.param


@pytest.fixture(scope='module')
def jwt_key_id():
    return 'jwt_key_id_1'


@pytest.fixture(scope='module')
def rsa_private_key_object():
    return generate_rsa_private_key(public_exponent=65537, key_size=4096, backend=default_backend())


@pytest.fixture(params=(None, b'strong_password'))
def rsa_passphrase(request):
    return request.param


@pytest.fixture
def rsa_private_key_bytes(rsa_private_key_object, rsa_passphrase):
    encryption = serialization.BestAvailableEncryption(rsa_passphrase) if rsa_passphrase else serialization.NoEncryption()
    return rsa_private_key_object.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=encryption,
    )


@pytest.fixture(scope='function')
def successful_token_response(successful_token_mock, successful_token_json_response):
    # pylint:disable=redefined-outer-name
    response = successful_token_json_response.copy()
    del response['refresh_token']
    successful_token_mock.json = Mock(return_value=response)
    successful_token_mock.ok = True
    successful_token_mock.content = json.dumps(response)
    successful_token_mock.status_code = 200
    return successful_token_mock


@pytest.mark.parametrize(('key_file', 'key_data'), [(None, None), ('fake sys path', 'fake key data')])
@pytest.mark.parametrize('rsa_passphrase', [None])
def test_jwt_auth_init_raises_type_error_unless_exactly_one_of_rsa_private_key_file_or_data_is_given(key_file, key_data, rsa_private_key_bytes):
    kwargs = dict(
        rsa_private_key_data=rsa_private_key_bytes,
        client_id=None,
        client_secret=None,
        jwt_key_id=None,
        enterprise_id=None,
    )
    JWTAuth(**kwargs)
    kwargs.update(rsa_private_key_file_sys_path=key_file, rsa_private_key_data=key_data)
    with pytest.raises(TypeError):
        JWTAuth(**kwargs)


@pytest.mark.parametrize('key_data', [object(), 'ƒøø'])
@pytest.mark.parametrize('rsa_passphrase', [None])
def test_jwt_auth_init_raises_type_error_if_rsa_private_key_data_has_unexpected_type(key_data, rsa_private_key_bytes):
    kwargs = dict(
        rsa_private_key_data=rsa_private_key_bytes,
        client_id=None,
        client_secret=None,
        jwt_key_id=None,
        enterprise_id=None,
    )
    JWTAuth(**kwargs)
    kwargs.update(rsa_private_key_data=key_data)
    with pytest.raises(TypeError):
        JWTAuth(**kwargs)


@pytest.mark.parametrize('rsa_private_key_data_type', [io.BytesIO, str, bytes, RSAPrivateKey])
def test_jwt_auth_init_accepts_rsa_private_key_data(rsa_private_key_bytes, rsa_passphrase, rsa_private_key_data_type):
    if rsa_private_key_data_type is str:
        rsa_private_key_data = str(rsa_private_key_bytes.decode('ascii'))
    elif rsa_private_key_data_type is RSAPrivateKey:
        rsa_private_key_data = serialization.load_pem_private_key(
            rsa_private_key_bytes,
            password=rsa_passphrase,
            backend=default_backend(),
        )
    else:
        rsa_private_key_data = rsa_private_key_data_type(rsa_private_key_bytes)
    JWTAuth(
        rsa_private_key_data=rsa_private_key_data,
        rsa_private_key_passphrase=rsa_passphrase,
        client_id=None,
        client_secret=None,
        jwt_key_id=None,
        enterprise_id=None,
    )


@pytest.fixture(params=[False, True])
def pass_private_key_by_path(request):
    """For jwt_auth_init_mocks, whether to pass the private key via sys_path (True) or pass the data directly (False)."""
    return request.param


@pytest.fixture
def jwt_auth_init_mocks(
        mock_box_session,
        successful_token_response,
        jwt_algorithm,
        jwt_key_id,
        rsa_passphrase,
        rsa_private_key_bytes,
        pass_private_key_by_path,
):
    # pylint:disable=redefined-outer-name

    @contextmanager
    def _jwt_auth_init_mocks(**kwargs):
        assert_authed = kwargs.pop('assert_authed', True)
        fake_client_id = 'fake_client_id'
        fake_client_secret = 'fake_client_secret'
        assertion = Mock()
        data = {
            'grant_type': JWTAuth._GRANT_TYPE,  # pylint:disable=protected-access
            'client_id': fake_client_id,
            'client_secret': fake_client_secret,
            'assertion': assertion,
            'box_device_id': '0',
            'box_device_name': 'my_awesome_device',
        }
        mock_box_session.request.return_value = successful_token_response
        with patch('boxsdk.auth.jwt_auth.open', mock_open(read_data=rsa_private_key_bytes), create=True) as jwt_auth_open:
            with patch('cryptography.hazmat.primitives.serialization.load_pem_private_key') as load_pem_private_key:
                oauth = JWTAuth(
                    client_id=fake_client_id,
                    client_secret=fake_client_secret,
                    rsa_private_key_file_sys_path=(sentinel.rsa_path if pass_private_key_by_path else None),
                    rsa_private_key_data=(None if pass_private_key_by_path else rsa_private_key_bytes),
                    rsa_private_key_passphrase=rsa_passphrase,
                    session=mock_box_session,
                    box_device_name='my_awesome_device',
                    jwt_algorithm=jwt_algorithm,
                    jwt_key_id=jwt_key_id,
                    enterprise_id=kwargs.pop('enterprise_id', None),
                    **kwargs
                )
                if pass_private_key_by_path:
                    jwt_auth_open.assert_called_once_with(sentinel.rsa_path, 'rb')
                    jwt_auth_open.return_value.read.assert_called_once_with()  # pylint:disable=no-member
                else:
                    jwt_auth_open.assert_not_called()
                load_pem_private_key.assert_called_once_with(
                    rsa_private_key_bytes,
                    password=rsa_passphrase,
                    backend=default_backend(),
                )

                yield oauth, assertion, fake_client_id, load_pem_private_key.return_value

        if assert_authed:
            mock_box_session.request.assert_called_once_with(
                'POST',
                f'{API.OAUTH2_API_URL}/token',
                data=data,
                headers={'content-type': 'application/x-www-form-urlencoded'},
                access_token=None,
            )
            assert oauth.access_token == successful_token_response.json()['access_token']

    return _jwt_auth_init_mocks


def test_refresh_authenticates_with_user_if_enterprise_id_and_user_both_passed_to_constructor(jwt_auth_init_and_auth_mocks):
    user = 'fake_user_id'
    with jwt_auth_init_and_auth_mocks(sub=user, sub_type='user', enterprise_id='fake_enterprise_id', user=user) as oauth:
        oauth.refresh(None)


@pytest.mark.parametrize('jwt_auth_method_name', ['authenticate_user', 'authenticate_instance'])
def test_authenticate_raises_value_error_if_sub_was_never_given(jwt_auth_init_mocks, jwt_auth_method_name):
    with jwt_auth_init_mocks(assert_authed=False) as params:
        auth = params[0]
        authenticate_method = getattr(auth, jwt_auth_method_name)
        with pytest.raises(ValueError):
            authenticate_method()


def test_jwt_auth_constructor_raises_type_error_if_user_is_unsupported_type(jwt_auth_init_mocks):
    with pytest.raises(TypeError):
        with jwt_auth_init_mocks(user=object()):
            assert False


def test_authenticate_user_raises_type_error_if_user_is_unsupported_type(jwt_auth_init_mocks):
    with jwt_auth_init_mocks(assert_authed=False) as params:
        auth = params[0]
        with pytest.raises(TypeError):
            auth.authenticate_user(object())


@pytest.mark.parametrize('user_id_for_init', [None, 'fake_user_id_1'])
def test_authenticate_user_saves_user_id_for_future_calls(jwt_auth_init_and_auth_mocks, user_id_for_init, jwt_encode):

    def assert_jwt_encode_call_args(user_id):
        assert jwt_encode.call_args[0][0]['sub'] == user_id
        assert jwt_encode.call_args[0][0]['box_sub_type'] == 'user'
        jwt_encode.call_args = None

    with jwt_auth_init_and_auth_mocks(sub=None, sub_type=None, assert_authed=False, user=user_id_for_init) as auth:
        for new_user_id in ['fake_user_id_2', 'fake_user_id_3']:
            auth.authenticate_user(new_user_id)
            assert_jwt_encode_call_args(new_user_id)
            auth.authenticate_user()
            assert_jwt_encode_call_args(new_user_id)


def test_authenticate_instance_raises_value_error_if_different_enterprise_id_is_given(jwt_auth_init_mocks):
    with jwt_auth_init_mocks(enterprise_id='fake_enterprise_id_1', assert_authed=False) as params:
        auth = params[0]
        with pytest.raises(ValueError):
            auth.authenticate_instance('fake_enterprise_id_2')


def test_authenticate_instance_saves_enterprise_id_for_future_calls(jwt_auth_init_and_auth_mocks):
    enterprise_id = 'fake_enterprise_id'
    with jwt_auth_init_and_auth_mocks(sub=enterprise_id, sub_type='enterprise', assert_authed=False) as auth:
        auth.authenticate_instance(enterprise_id)
        auth.authenticate_instance()
        auth.authenticate_instance(enterprise_id)
        with pytest.raises(ValueError):
            auth.authenticate_instance('fake_enterprise_id_2')


@pytest.fixture
def jwt_encode():
    with patch('jwt.encode') as patched_jwt_encode:
        yield patched_jwt_encode


@pytest.fixture
def jwt_auth_auth_mocks(jti_length, jwt_algorithm, jwt_key_id, jwt_encode):

    @contextmanager
    def _jwt_auth_auth_mocks(sub, sub_type, oauth, assertion, client_id, secret, assert_authed=True):
        # pylint:disable=redefined-outer-name
        with patch('boxsdk.auth.jwt_auth.datetime') as mock_datetime:
            with patch('boxsdk.auth.jwt_auth.random.SystemRandom') as mock_system_random:
                jwt_encode.return_value = assertion
                mock_datetime.utcnow.return_value = datetime(2015, 7, 6, 12, 1, 2)
                mock_datetime.return_value = datetime(1970, 1, 1)
                now_plus_30 = mock_datetime.utcnow.return_value + timedelta(seconds=30)
                exp = int((now_plus_30 - datetime(1970, 1, 1)).total_seconds())
                system_random = mock_system_random.return_value
                system_random.randint.return_value = jti_length
                random_choices = [random.random() for _ in range(jti_length)]

                # Use cycle so that we can do auth more than once inside the context manager.
                system_random.random.side_effect = cycle(random_choices)

                ascii_alphabet = string.ascii_letters + string.digits
                ascii_len = len(ascii_alphabet)
                jti = ''.join(ascii_alphabet[int(r * ascii_len)] for r in random_choices)

                yield oauth

                if assert_authed:
                    system_random.randint.assert_called_once_with(16, 128)
                    assert len(system_random.random.mock_calls) == jti_length
                    jwt_encode.assert_called_once_with({
                        'iss': client_id,
                        'sub': sub,
                        'box_sub_type': sub_type,
                        'aud': 'https://api.box.com/oauth2/token',
                        'jti': jti,
                        'exp': exp,
                    }, secret, algorithm=jwt_algorithm, headers={'kid': jwt_key_id})

    return _jwt_auth_auth_mocks


@pytest.fixture
def jwt_auth_init_and_auth_mocks(jwt_auth_init_mocks, jwt_auth_auth_mocks):

    @contextmanager
    def _jwt_auth_init_and_auth_mocks(sub, sub_type, *jwt_auth_init_mocks_args, **jwt_auth_init_mocks_kwargs):
        assert_authed = jwt_auth_init_mocks_kwargs.pop('assert_authed', True)
        with jwt_auth_init_mocks(*jwt_auth_init_mocks_args, assert_authed=assert_authed, **jwt_auth_init_mocks_kwargs) as params:
            with jwt_auth_auth_mocks(sub, sub_type, *params, assert_authed=assert_authed) as oauth:
                yield oauth

    return _jwt_auth_init_and_auth_mocks


@pytest.mark.parametrize(
    ('user', 'pass_in_init'),
    list(product([str('fake_user_id'), User(None, 'fake_user_id')], [False, True])),
)
def test_authenticate_user_sends_post_request_with_correct_params(jwt_auth_init_and_auth_mocks, user, pass_in_init):
    # pylint:disable=redefined-outer-name
    if isinstance(user, User):
        user_id = user.object_id
    elif isinstance(user, str):
        user_id = user
    else:
        raise NotImplementedError
    init_kwargs = {}
    authenticate_params = []
    if pass_in_init:
        init_kwargs['user'] = user
    else:
        authenticate_params.append(user)
    with jwt_auth_init_and_auth_mocks(user_id, 'user', **init_kwargs) as oauth:
        oauth.authenticate_user(*authenticate_params)


@pytest.mark.parametrize(('pass_in_init', 'pass_in_auth'), [(True, False), (False, True), (True, True)])
def test_authenticate_instance_sends_post_request_with_correct_params(jwt_auth_init_and_auth_mocks, pass_in_init, pass_in_auth):
    # pylint:disable=redefined-outer-name
    enterprise_id = 'fake_enterprise_id'
    init_kwargs = {}
    auth_params = []
    if pass_in_init:
        init_kwargs['enterprise_id'] = enterprise_id
    if pass_in_auth:
        auth_params.append(enterprise_id)
    with jwt_auth_init_and_auth_mocks(enterprise_id, 'enterprise', **init_kwargs) as oauth:
        oauth.authenticate_instance(*auth_params)


def test_refresh_app_user_sends_post_request_with_correct_params(jwt_auth_init_and_auth_mocks):
    # pylint:disable=redefined-outer-name
    fake_user_id = 'fake_user_id'
    with jwt_auth_init_and_auth_mocks(fake_user_id, 'user', user=fake_user_id) as oauth:
        oauth.refresh(None)


def test_refresh_instance_sends_post_request_with_correct_params(jwt_auth_init_and_auth_mocks):
    # pylint:disable=redefined-outer-name
    enterprise_id = 'fake_enterprise_id'
    with jwt_auth_init_and_auth_mocks(enterprise_id, 'enterprise', enterprise_id=enterprise_id) as oauth:
        oauth.refresh(None)


@pytest.fixture()
def jwt_subclass_that_just_stores_params():
    class StoreParamJWTAuth(JWTAuth):
        def __init__(self, **kwargs):
            self.kwargs = kwargs
            super().__init__(**kwargs)

    return StoreParamJWTAuth


@pytest.fixture
def fake_client_id():
    return 'fake_client_id'


@pytest.fixture
def fake_client_secret():
    return 'fake_client_secret'


@pytest.fixture
def fake_enterprise_id():
    return 'fake_enterprise_id'


@pytest.fixture
def app_config_json_content(
        fake_client_id,
        fake_client_secret,
        fake_enterprise_id,
        jwt_key_id,
        rsa_private_key_bytes,
        rsa_passphrase,
):
    template = r"""
{{
  "boxAppSettings": {{
    "clientID": "{client_id}",
    "clientSecret": "{client_secret}",
    "appAuth": {{
      "publicKeyID": "{jwt_key_id}",
      "privateKey": "{private_key}",
      "passphrase": {passphrase}
    }}
  }},
  "enterpriseID": {enterprise_id}
}}"""
    return template.format(
        client_id=fake_client_id,
        client_secret=fake_client_secret,
        jwt_key_id=jwt_key_id,
        private_key=rsa_private_key_bytes.replace(b"\n", b"\\n").decode(),
        passphrase=json.dumps(rsa_passphrase and rsa_passphrase.decode()),
        enterprise_id=json.dumps(fake_enterprise_id),
    )


@pytest.fixture()
def assert_jwt_kwargs_expected(
        fake_client_id,
        fake_client_secret,
        fake_enterprise_id,
        jwt_key_id,
        rsa_private_key_bytes,
        rsa_passphrase,
):
    def _assert_jwt_kwargs_expected(jwt_auth):
        assert jwt_auth.kwargs['client_id'] == fake_client_id
        assert jwt_auth.kwargs['client_secret'] == fake_client_secret
        assert jwt_auth.kwargs['enterprise_id'] == fake_enterprise_id
        assert jwt_auth.kwargs['jwt_key_id'] == jwt_key_id
        assert jwt_auth.kwargs['rsa_private_key_data'] == rsa_private_key_bytes.decode()
        assert jwt_auth.kwargs['rsa_private_key_passphrase'] == (rsa_passphrase and rsa_passphrase.decode())

    return _assert_jwt_kwargs_expected


def test_from_config_file(
        jwt_subclass_that_just_stores_params,
        app_config_json_content,
        assert_jwt_kwargs_expected,
):
    # pylint:disable=redefined-outer-name
    with patch('boxsdk.auth.jwt_auth.open', mock_open(read_data=app_config_json_content), create=True):
        jwt_auth_from_config_file = jwt_subclass_that_just_stores_params.from_settings_file('fake_config_file_sys_path')
        assert_jwt_kwargs_expected(jwt_auth_from_config_file)


def test_from_settings_dictionary(
        jwt_subclass_that_just_stores_params,
        app_config_json_content,
        assert_jwt_kwargs_expected,
):
    jwt_auth_from_dictionary = jwt_subclass_that_just_stores_params.from_settings_dictionary(json.loads(app_config_json_content))
    assert_jwt_kwargs_expected(jwt_auth_from_dictionary)


@pytest.fixture
def expect_auth_retry(status_code, error_description, include_date_header, error_code):
    return status_code == 400 and ('exp' in error_description or 'jti' in error_description) and include_date_header and error_code == 'invalid_grant'


@pytest.fixture
def box_datetime():
    return datetime.now(tz=pytz.utc) - timedelta(100)


@pytest.fixture
def unsuccessful_jwt_response(box_datetime, status_code, error_description, include_date_header, error_code):
    headers = {'Date': box_datetime.strftime('%a, %d %b %Y %H:%M:%S %Z')} if include_date_header else {}
    unsuccessful_response = Mock(headers=headers)
    unsuccessful_response.json.return_value = {'error_description': error_description, 'error': error_code}
    unsuccessful_response.status_code = status_code
    unsuccessful_response.ok = False
    return unsuccessful_response


@pytest.mark.parametrize('jwt_algorithm', ('RS512',))
@pytest.mark.parametrize('rsa_passphrase', (None,))
@pytest.mark.parametrize('pass_private_key_by_path', (False,))
@pytest.mark.parametrize('status_code', (400, 401))
@pytest.mark.parametrize('error_description', (
    'invalid box_sub_type claim', 'invalid kid', "check the 'exp' claim", "check the 'jti' claim"))
@pytest.mark.parametrize('error_code', ('invalid_grant', 'bad_request'))
@pytest.mark.parametrize('include_date_header', (True, False))
def test_auth_retry_for_invalid_exp_claim(
        jwt_auth_init_mocks,
        expect_auth_retry,
        unsuccessful_jwt_response,
        box_datetime,
):
    # pylint:disable=redefined-outer-name
    enterprise_id = 'fake_enterprise_id'
    with jwt_auth_init_mocks(assert_authed=False) as params:
        auth = params[0]
        with patch.object(auth, '_fetch_access_token') as mock_send_jwt:
            mock_send_jwt.side_effect = [BoxOAuthException(400, network_response=unsuccessful_jwt_response), 'jwt_token']
            if not expect_auth_retry:
                with pytest.raises(BoxOAuthException):
                    auth.authenticate_instance(enterprise_id)
            else:
                auth.authenticate_instance(enterprise_id)
            expected_calls = [call(enterprise_id, 'enterprise', None)]
            if expect_auth_retry:
                expected_calls.append(call(enterprise_id, 'enterprise', box_datetime.replace(microsecond=0, tzinfo=None)))
            assert len(mock_send_jwt.mock_calls) == len(expected_calls)
            mock_send_jwt.assert_has_calls(expected_calls)


@pytest.mark.parametrize('jwt_algorithm', ('RS512',))
@pytest.mark.parametrize('rsa_passphrase', (None,))
@pytest.mark.parametrize('pass_private_key_by_path', (False,))
@pytest.mark.parametrize('status_code', (429,))
@pytest.mark.parametrize('error_description', ('Request rate limit exceeded',))
@pytest.mark.parametrize('error_code', ('rate_limit_exceeded',))
@pytest.mark.parametrize('include_date_header', (False,))
def test_auth_retry_for_rate_limit_error(
        jwt_auth_init_mocks,
        unsuccessful_jwt_response,
):
    # pylint:disable=redefined-outer-name
    enterprise_id = 'fake_enterprise_id'
    with jwt_auth_init_mocks(assert_authed=False) as params:
        auth = params[0]
        with patch.object(auth, '_fetch_access_token') as mock_send_jwt:
            side_effect = []
            expected_calls = []
            # Retries multiple times, but less than max retries. Then succeeds when it gets a token.
            for _ in range(API.MAX_RETRY_ATTEMPTS - 2):
                side_effect.append(BoxOAuthException(429, network_response=unsuccessful_jwt_response))
                expected_calls.append(call(enterprise_id, 'enterprise', None))
            side_effect.append('jwt_token')
            expected_calls.append(call(enterprise_id, 'enterprise', None))
            mock_send_jwt.side_effect = side_effect

            auth.authenticate_instance(enterprise_id)
            assert len(mock_send_jwt.mock_calls) == len(expected_calls)
            mock_send_jwt.assert_has_calls(expected_calls)


@pytest.mark.parametrize('jwt_algorithm', ('RS512',))
@pytest.mark.parametrize('rsa_passphrase', (None,))
@pytest.mark.parametrize('pass_private_key_by_path', (False,))
@pytest.mark.parametrize('status_code', (429,))
@pytest.mark.parametrize('error_description', ('Request rate limit exceeded',))
@pytest.mark.parametrize('error_code', ('rate_limit_exceeded',))
@pytest.mark.parametrize('include_date_header', (False,))
def test_auth_max_retries_for_rate_limit_error(
        jwt_auth_init_mocks,
        unsuccessful_jwt_response,
):
    # pylint:disable=redefined-outer-name
    enterprise_id = 'fake_enterprise_id'
    with jwt_auth_init_mocks(assert_authed=False) as params:
        auth = params[0]
        with patch.object(auth, '_fetch_access_token') as mock_send_jwt:
            side_effect = []
            expected_calls = []
            # Retries max number of times, then throws the error
            for _ in range(API.MAX_RETRY_ATTEMPTS + 1):
                side_effect.append(BoxOAuthException(429, network_response=unsuccessful_jwt_response))
                expected_calls.append(call(enterprise_id, 'enterprise', None))
            mock_send_jwt.side_effect = side_effect

            with pytest.raises(BoxOAuthException) as error:
                auth.authenticate_instance(enterprise_id)
            assert error.value.status == 429
            assert len(mock_send_jwt.mock_calls) == len(expected_calls)
            mock_send_jwt.assert_has_calls(expected_calls)


@pytest.mark.parametrize('jwt_algorithm', ('RS512',))
@pytest.mark.parametrize('rsa_passphrase', (None,))
@pytest.mark.parametrize('pass_private_key_by_path', (False,))
@pytest.mark.parametrize('status_code', (500,))
@pytest.mark.parametrize('error_description', ('Internal Server Error',))
@pytest.mark.parametrize('error_code', ('internal_server_error',))
@pytest.mark.parametrize('include_date_header', (False,))
def test_auth_retry_for_internal_server_error(
        jwt_auth_init_mocks,
        unsuccessful_jwt_response,
):
    # pylint:disable=redefined-outer-name
    enterprise_id = 'fake_enterprise_id'
    with jwt_auth_init_mocks(assert_authed=False) as params:
        auth = params[0]
        with patch.object(auth, '_fetch_access_token') as mock_send_jwt:
            side_effect = []
            expected_calls = []
            # Retries multiple times, but less than max retries. Then succeeds when it gets a token.
            for _ in range(API.MAX_RETRY_ATTEMPTS - 2):
                side_effect.append(BoxOAuthException(500, network_response=unsuccessful_jwt_response))
                expected_calls.append(call(enterprise_id, 'enterprise', None))
            side_effect.append('jwt_token')
            expected_calls.append(call(enterprise_id, 'enterprise', None))
            mock_send_jwt.side_effect = side_effect

            auth.authenticate_instance(enterprise_id)
            assert len(mock_send_jwt.mock_calls) == len(expected_calls)
            mock_send_jwt.assert_has_calls(expected_calls)


@pytest.mark.parametrize('jwt_algorithm', ('RS512',))
@pytest.mark.parametrize('rsa_passphrase', (None,))
@pytest.mark.parametrize('pass_private_key_by_path', (False,))
@pytest.mark.parametrize('status_code', (500,))
@pytest.mark.parametrize('error_description', ('Internal Server Error',))
@pytest.mark.parametrize('error_code', ('internal_server_error',))
@pytest.mark.parametrize('include_date_header', (False,))
def test_auth_max_retries_for_internal_server_error(
        jwt_auth_init_mocks,
        unsuccessful_jwt_response,
):
    # pylint:disable=redefined-outer-name
    enterprise_id = 'fake_enterprise_id'
    with jwt_auth_init_mocks(assert_authed=False) as params:
        auth = params[0]
        with patch.object(auth, '_fetch_access_token') as mock_send_jwt:
            side_effect = []
            expected_calls = []
            # Retries max number of times, then throws the error
            for _ in range(API.MAX_RETRY_ATTEMPTS + 1):
                side_effect.append(BoxOAuthException(500, network_response=unsuccessful_jwt_response))
                expected_calls.append(call(enterprise_id, 'enterprise', None))
            mock_send_jwt.side_effect = side_effect

            with pytest.raises(BoxOAuthException) as error:
                auth.authenticate_instance(enterprise_id)
            assert error.value.status == 500
            assert len(mock_send_jwt.mock_calls) == len(expected_calls)
            mock_send_jwt.assert_has_calls(expected_calls)



import json

from boxsdk.object.comment import Comment


# pylint:disable=protected-access
# pylint:disable=redefined-outer-name

def test_reply(test_comment, mock_box_session, comment_params):
    expected_url = mock_box_session.get_url('comments')
    message_type, message = comment_params
    expected_data = {
        message_type: message,
        'item': {
            'type': 'comment',
            'id': test_comment.object_id
        }
    }
    mock_box_session.post.return_value.json.return_value = {
        'type': 'comment',
        'id': '12345',
        message_type: message
    }
    reply_comment = test_comment.reply(message)
    mock_box_session.post.assert_called_once_with(expected_url, data=json.dumps(expected_data))
    assert isinstance(reply_comment, Comment)
    assert reply_comment.object_id == '12345'


def test_edit(test_comment, mock_box_session, comment_params):
    expected_url = mock_box_session.get_url('comments', test_comment.object_id)
    message_type, message = comment_params
    expected_data = {
        message_type: message,
    }
    mock_box_session.put.return_value.json.return_value = {
        'type': 'comment',
        'id': '12345',
        message_type: message
    }
    updated_comment = test_comment.edit(message)
    mock_box_session.put.assert_called_once_with(expected_url, data=json.dumps(expected_data), headers=None, params=None)
    assert isinstance(updated_comment, Comment)
    assert updated_comment[message_type] == message


def test_get(mock_box_session):
    comment_id = '1235'
    expected_url = mock_box_session.get_url('comments', comment_id)
    mock_box_session.get.return_value.json.return_value = {
        'type': 'comment',
        'id': comment_id,
        'message': 'Hi!'
    }
    comment = Comment(mock_box_session, comment_id).get()
    mock_box_session.get.assert_called_once_with(expected_url, headers=None, params=None)
    assert comment.object_id == comment_id
    assert comment.message == 'Hi!'  # pylint:disable=no-member


def test_delete(test_comment, mock_box_session):
    expected_url = test_comment.get_url()
    test_comment.delete()
    mock_box_session.delete.assert_called_once_with(expected_url, expect_json_response=False, headers=None, params={})



from collections import ChainMap
import inspect

__all__ = list(map(str, ['Translator']))

from typing import Any, TYPE_CHECKING, Dict

if TYPE_CHECKING:
    from boxsdk.object.base_api_json_object import BaseAPIJSONObjectMeta
    from boxsdk.session.session import Session


def _get_object_id(obj: dict) -> Any:
    """
    Gets the ID for an API object.

    :param obj:
        The API object
    :return:
    """
    return obj.get('id', None)


class Translator(ChainMap):
    """
    Translate item responses from the Box API to Box objects.

    Also acts as a :class:`Mapping` from type names to Box object classes.

    There exists a global default `Translator`, containing the default API
    object classes defined by the SDK. Custom `Translator` instances can be
    created to extend the default `Translator` with custom subclasses.

    A `Translator` is a :class:`ChainMap`, so that one translator can "extend"
    others. The most common scenario would be a custom, non-global
    `Translator` that extends only the default translator, to register 0 or
    more new classes. But more complex inheritance is also allowed, in case
    that is useful.
    """

    __slots__ = ()

    # :attr _default_translator:
    #     A global `Translator` containing the default API object classes
    #     defined by the SDK. By default, new `Translator` instances will
    #     "extend" this one, so that the global registrations are reflected
    #     automatically.
    #
    #     NOTE: For convenience and backwards-compatability, developers are
    #     allowed to register their own custom subclasses with
    #     `_default_translator`, but are encouraged not to. The default
    #     translator may change or be removed in any major or minor release.
    #     Additionally, it has the usual hazards of mutable global state.
    #     The supported and recommended ways for registering custom subclasses
    #     are:
    #
    #     - Constructing a new `Translator`, calling `Translator.register()` as
    #     necessary, and passing it to the `BoxSession` constructor.
    #     - Calling `session.translator.register()` on an existing
    #     `BoxSession`.
    #     - Calling `client.translator.register()` on an existing `Client`.
    # :type _default_translator:   :class:`Translator`
    _default_translator = {}   # Will be set to a `Translator` instance below, after the class is defined.

    def __init__(self, *translation_maps: Dict[str, 'BaseAPIJSONObjectMeta'], **kwargs: Any):
        """Baseclass override.

        :param translation_maps:
            (variadic) The same as the `maps` variadic parameter to
            :class:`ChainMap`, except restricted to maps from type names to Box
            object classes.
        :param extend_default_translator:
            (optional, keyword-only) If `True` (the default),
            `_default_translator` is appended to the end of `translation_maps`.
            When this functionality is used, the new `Translator` will inherit
            all of the global registrations.
        :type extend_default_translator:  `bool`
        :param new_child:
            (optional, keyword-only) If `True` (the default), a new empty
            `dict` is prepended to the front of `translation_maps`. Either way,
            the resulting `Translator` starts out with the same key-value
            pairs. But when this is `False`, the first item in
            `translation_maps` will be mutated by `__setitem__()` and
            `__delitem()__` calls, which will affect other references to it.
            Whereas when this is `True`, all items in `translation_maps` are
            safe from mutation in normal usage scenarios.
        :type new_child:  `bool`
        """
        translation_maps = list(translation_maps)
        extend_default_translator = kwargs.pop('extend_default_translator', True)
        new_child = kwargs.pop('new_child', True)
        if extend_default_translator:
            translation_maps.append(self._default_translator)
        if new_child:
            translation_maps.insert(0, {})
        super().__init__(*translation_maps, **kwargs)

    def register(self, type_name: str, box_cls: 'BaseAPIJSONObjectMeta') -> Any:
        """Associate a Box object class to handle Box API item responses with the given type name.

        :param type_name:
            The type name to be registered.
        :param box_cls:
            The Box object class, which will be associated with the type name provided.
        """
        self[type_name] = box_cls

    def get(self, key: str, default: 'BaseAPIJSONObjectMeta' = None) -> 'BaseAPIJSONObjectMeta':
        """Get the box object class associated with the given type name.

        :param key:
            The type name to be translated.
        :param default:
            (optional) The default Box object class to return.
            Defaults to `BaseObject`.
        """
        # pylint:disable=import-outside-toplevel
        from boxsdk.object.base_object import BaseObject
        if default is None:
            default = BaseObject
        return super().get(key, default)

    def translate(self, session: 'Session', response_object: dict) -> Any:
        """
        Translate a given API response object into SDK classes, rescursively translating any subobjects.

        :param session:
            The SDK session to use for any objects that require a session (i.e. classes that make API calls)
        :param response_object:
            The JSON response object from the API, which will be translated
        :return:
            The translated object
        """

        if not isinstance(response_object, dict):
            return response_object

        translated_obj = {}
        object_type = response_object.get('type', None)
        object_class = self.get(object_type) if object_type is not None else None
        # Parent classes have the ability to skip fields that they do not want translated
        fields_to_skip = object_class.untranslated_fields() if object_class is not None else ()
        for key in response_object:
            if key in fields_to_skip:
                translated_obj[key] = response_object[key]
                continue
            if isinstance(response_object[key], dict):
                translated_obj[key] = self.translate(session, response_object[key])
            elif isinstance(response_object[key], list):
                translated_obj[key] = [self.translate(session, o) for o in response_object[key]]
            else:
                translated_obj[key] = response_object[key]

        # Try to translate any API object with a `type` property, except for metadata instances
        # The $type value in metadata instances isn't directly usable, so we avoid it altogether
        # NOTE: Currently, we represent metadata as just a `dict`, so there's no need to translate it anyway
        # Metadata field objects are another issue; they contain a 'type' property that doesn't really
        # map to a Box object.  We probably want to treat these as just `dict`s, so they're excluded here
        if object_class is not None and '$type' not in translated_obj:
            param_values = {
                'session': session,
                'response_object': translated_obj,
                'object_id': _get_object_id(translated_obj),
            }
            params = inspect.signature(object_class.__init__).parameters
            param_values = {p: param_values[p] for p in params if p in param_values}
            return object_class(**param_values)

        return translated_obj


Translator._default_translator = Translator(extend_default_translator=False)  # pylint:disable=protected-access



from enum import Enum


class TextEnum(str, Enum):
    def __repr__(self):
        return self._value_  # pylint:disable=no-member

    def __str__(self):
        return str(self.value)  # pylint:disable=no-member



import json
import pytest

from boxsdk.config import API
from boxsdk.object.collaboration_allowlist_entry import CollaborationAllowlistEntry
from boxsdk.object.collaboration_allowlist_exempt_target import CollaborationAllowlistExemptTarget


def test_get_entries(mock_box_session, test_collaboration_allowlist):
    expected_url = f'{API.BASE_API_URL}/collaboration_whitelist_entries'
    mock_entry = {
        'type': 'collaboration_whitelist_entry',
        'id': '12345',
        'domain': 'box.com',
        'direction': 'both'
    }
    mock_box_session.get.return_value.json.return_value = {
        'limit': 100,
        'entries': [mock_entry]
    }
    entries = test_collaboration_allowlist.get_entries()
    entry = entries.next()
    mock_box_session.get.assert_called_once_with(expected_url, params={})
    assert isinstance(entry, CollaborationAllowlistEntry)
    assert entry.id == mock_entry['id']
    assert entry.direction == mock_entry['direction']
    assert entry.domain == mock_entry['domain']


def test_get_exemptions(mock_box_session, test_collaboration_allowlist):
    expected_url = f'{API.BASE_API_URL}/collaboration_whitelist_exempt_targets'
    mock_exemption = {
        'type': 'collaboration_whitelist_exempt_target',
        'id': '12345',
        'user': {
            'type': 'user',
            'id': '33333'
        }
    }
    mock_box_session.get.return_value.json.return_value = {
        'limit': 100,
        'entries': [mock_exemption]
    }
    exemptions = test_collaboration_allowlist.get_exemptions()
    exemption = exemptions.next()
    mock_box_session.get.assert_called_once_with(expected_url, params={})
    assert isinstance(exemption, CollaborationAllowlistExemptTarget)
    assert exemption.id == mock_exemption['id']
    assert exemption.user['id'] == mock_exemption['user']['id']


@pytest.mark.parametrize(
    'direction',
    ['inbound', 'outbound', 'both']
)
def test_add_domain(mock_box_session, test_collaboration_allowlist, direction):
    expected_url = f'{API.BASE_API_URL}/collaboration_whitelist_entries'
    domain = 'example.com'
    mock_entry = {
        'type': 'collaboration_whitelist_entry',
        'id': '12345',
        'domain': domain,
        'direction': direction
    }
    expected_data = {
        'domain': domain,
        'direction': direction
    }
    mock_box_session.post.return_value.json.return_value = mock_entry
    entry = test_collaboration_allowlist.add_domain(domain, direction)
    mock_box_session.post.assert_called_once_with(expected_url, data=json.dumps(expected_data))
    assert isinstance(entry, CollaborationAllowlistEntry)
    assert entry.id == mock_entry['id']
    assert entry.domain == domain
    assert entry.direction == direction


def test_add_exemption(mock_box_session, test_collaboration_allowlist, mock_user):
    expected_url = f'{API.BASE_API_URL}/collaboration_whitelist_exempt_targets'
    expected_data = {
        'user': {
            'id': mock_user.object_id
        }
    }
    mock_exemption = {
        'type': 'collaboration_whitelist_exempt_target',
        'id': '12345',
        'user': {
            'type': 'user',
            'id': mock_user.object_id
        }
    }
    mock_box_session.post.return_value.json.return_value = mock_exemption
    exemption = test_collaboration_allowlist.add_exemption(mock_user)
    mock_box_session.post.assert_called_once_with(expected_url, data=json.dumps(expected_data))
    assert exemption.id == mock_exemption['id']
    assert exemption.user['id'] == mock_exemption['user']['id']
